/***************************************************************************
 *                                                                         *
 *  Copyright 1995 Symbios Logic Incorporated.  All rights reserved.       *
 *                                                                         *
 * This software was written by Symbios Logic Inc. to develop and test new *
 * products.  Symbios Logic assumes no liability for its use. This software*
 * is released to the public domain to illustrate certain programming      *
 * techniques for testing the 53c8xx chips                                 *
 *                                                                         *
 ***************************************************************************/

/*           Name:  8XX_SCSI.C
 *          Title:  Validation test code 53C8XX SCSI tools
 *      $Workfile:  $
 *      $Revision:  $
 *       $Modtime:  $
 *     Programmer:  Symbios Logic
 *  Creation Date:  11/29/95
 *
 *  Version History
 *  ---------------
 *
 *    Date    Who?  Description
 *  --------  ----  -------------------------------------------------------
#BeginRevision
 *  03/06/96  SYM   - Made general formatting changes and documentation
 *						  additions
 *						  - Changed do_syncneg so that C8XXDevice structure is
 *						  used instead of explicitly checking for devices.  The
 *						  function will now return NULL if the C8XX device is not
 *						  recognized.
 *						  - Fixed problem in do_syncneg in which we were looking
 *						  at too many bits when determining if there was an
 *						  80 MHz clock attached
 *						  - Changed init_scsi such that the chip must support
 *						  wide scsi to negotiate for wide scsi
 *						  - Added auto request sense feature to do_rw function
 *						  - Changed printf's of inquiry data in init_scsi to
 *						  calls to writelog.  Also changed format of inquiry print
 *						  out.  Added a device number field along with wide, sync
 *						  and tcq capability indication.
 *						  - Fixed problem in init_scsi in which the SCSIDevice
 *						  linked list was not being terminated correctly with
 *						  a NULL pointer.
 *						  - Added SCSI Bus reset to init_scsi function
 *						  - Fixed a problem with selecting narrow devices after a
 *						  wide device was selected in the init_scsi routine.  The
 *						  wide bit in the select TI entry was being carried over
 *						  to the next device.
 *						  - Added clear DMA fifo to the end of data_out phase
 *						  mismatch code in the do_rw function.
 *						  - Added code in init_scsi to skip over scanning the chip's
 *						  scsi id during the bus scan.
 *						  - Added do_StartStopUnit scsi function.
 *						  - Added auto Start Unit to init_scsi
 *						  - Removed else at end of do_rw
 *						  - Reworked do_RqSense such that the Int handling algorithm
 *						  is more efficient
 *						  - Move Data In/Data Out phase mismatch handling code out of
 *						  do_rw and into separate local functions.
 *
#EndRevision
 *
 *  11/29/95  SYM  Initial Revision.
 *
#BeginDescription
 *
 * This file contains functions for executing common scsi sequences using
 * 53C8XX devices.  Sequences include Test Unit Ready, Request Sense,
 * Inquiry, Read Capacity, Start/Stop Unit Synchronous Negotiation, Wide Neg.,
 * 6 byte Read, 10 byte Read, 6 byte Write, and 10 byte Write.  There is
 * also a general SCSI initialization routine that can be used to find
 * and initialize devices on the scsi bus as well as negotiate synchronous
 * and wide.  These functions are hardware specific to the 53C8XX family
 * of devices and would have to be rather dramatically changed to be made
 * to work with other devices.
 *
#EndDescription
 *
 *-------------------------------------------------------------------------
 *
 *  $Header: $
 *
 */

#include <dos.h>
#include <alloc.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "valtypes.h"
#include "maint_sc.h"           /* maintenance script, for tur etc. */
#include "rw_scr.h"             /* read write script, for reads and writes */
/* the following two #includes must go in this order */
#include "pci.h"
#include "gen_tool.h"
#include "8xx_pci.h"
#include "8xx_tool.h"
#include "8xx_scsi.h"

VINT init_gpscripts(void);
scsidev_record *init_scsi(pcidev_record *PCIDevice, VINT do_sync, VINT do_wide);
VINT do_TUR(pcidev_record *PCIDevice, scsimaint *maint);
VINT do_RqSense(pcidev_record *PCIDevice, scsimaint *maint);
VINT do_Inquiry(pcidev_record *PCIDevice, scsimaint *maint);
VINT do_RdCap(pcidev_record *PCIDevice, scsimaint *maint);
VINT do_StartStopUnit(pcidev_record *PCIDevice, scsimaint *maint);
VINT do_SyncNeg(pcidev_record *PCIDevice, scsimaint *maint);
VINT do_WideNeg(pcidev_record *PCIDevice, scsimaint *maint);
VINT do_Read6(pcidev_record *PCIDevice, scsirw *rw);
VINT do_Read10(pcidev_record *PCIDevice, scsirw *rw);
VINT do_Write6(pcidev_record *PCIDevice, scsirw *rw);
VINT do_Write10(pcidev_record *PCIDevice, scsirw *rw);

/* local functions */
static VINT do_rw(pcidev_record *PCIDevice, scsirw *rw);
static void Fill_SG_Entries(pcidev_record *PCIDevice, scsirw *rw, VINT *NextSGEntry);
static void HandleDataOutPM(pcidev_record *PCIDevice);
static void HandleDataInPM(pcidev_record *PCIDevice, VINT *Current_SG_Entry);
static VINT Bitbucket(pcidev_record *PCIDevice);
static VINT Fail_SCSI(pcidev_record *PCIDevice, VUBYTE istat, VUBYTE dstat,
				 VUINT sist, VULONG vector, VINT fail_type);



/* local variables */
static VULONG far *script;              /* ptr to maintenance scripts */
static VULONG far *rw_script;           /* ptr to rw scripts */
static ti_entry far *maint_buffer_table;/* ptr to maintenance ti table */
static ti_entry far *rw_buffer_table;   /* ptr to read/write ti table */
static VUBYTE Rqsns_or_Inquiry;         /* Used to tell whether to do Rqsense or Inquiry */
static VUBYTE msg_reject_buf;           /* general buffer to hold msg reject */
													 /* on rejected negotiation */
static VUBYTE bitbucket;                /* holds trash data on failing io */
static VULONG FirstDIMove_paddr;        /* Used for calculating where in the */
													 /* scatter gather list we are when a */
													 /* phase mismatch occurs */
static VULONG FirstDOMove_paddr;        /* same as above except for data out */
													 /* transfers */
static VUBYTE buffer[0xFF];             /* generic buffer for data transfers */
static VCHAR string[80];                /* generic string buffer for printing */

/*  The following table was created to simplify the task of 'computing' divisor
 *  values for various synchronous communication speeds at the only two clock
 *  settings we have to handle.
 *
 *  The first value in the table is the requested period as it would come
 *  across the SCSI bus.  The next two columns store the SCNTL3 and SXFER
 *  register values to program for the given speed.
 */

static VUBYTE  Sync80Table[][3] =
/*                                              Target      Ours */
/* Period, Scntl, Sxfer         Scntl & Sxfer   Rate    Read    Write */
	 { {13,  0x15, 0x00},     /*   1   &   4     19.23   20.00   20.00 */
		{15,  0x15, 0x20},     /*   1   &   5     16.67   20.00   16.00 */
		{18,  0x15, 0x40},     /*   1   &   6     13.89   20.00   13.33 */
		{24,  0x15, 0x80},     /*   1   &   8     10.41   20.00   10.00 */
		{25,  0x35, 0x00},     /*   2   &   4     10.00   10.00   10.00 */
		{31,  0x35, 0x20},     /*   2   &   5      8.06   10.00    8.00 */
		{43,  0x35, 0x60},     /*   2   &   7      5.81   10.00    5.71 */
		{49,  0x35, 0x80},     /*   2   &   8      5.10   10.00    5.00 */
		{50,  0x55, 0x00},     /*   4   &   4      5.00    5.00    5.00 */
		{62,  0x55, 0x20},     /*   4   &   5      4.03    5.00    4.00 */
		{75,  0x55, 0x40},     /*   4   &   6      3.33    5.00    3.33 */
		{84,  0x45, 0xA0},     /*   3   &   9      2.98    6.67    2.96 */
		{255, 0x55, 0xE0}      /*   4   &   11     0.00    5.00    1.81 */
	 };

static VUBYTE  Sync40Table[][3] =
/*                                              Target      Ours */
/* Period, Scntl, Sxfer         Scntl & Sxfer   Rate    Read    Write */
	 { {25,  0x13, 0x00},     /*   1   &   4     10.00   10.00   10.00 */
		{31,  0x13, 0x20},     /*   1   &   5      8.06   10.00    8.00 */
		{37,  0x13, 0x40},     /*   1   &   6      6.75   10.00    6.67 */
		{43,  0x13, 0x60},     /*   1   &   7      5.81   10.00    5.71 */
		{49,  0x13, 0x80},     /*   1   &   8      5.10   10.00    5.00 */
		{50,  0x33, 0x00},     /*   2   &   4      5.00    5.00    5.00 */
		{62,  0x33, 0x20},     /*   2   &   5      4.03    5.00    4.00 */
		{100, 0x33, 0x80},     /*   2   &   8      2.50    5.00    2.50 */
		{255, 0x43, 0xE0}      /*   3   &   11     0.00    3.33    1.21 */
	 };


/******************************************************************************
 Function: init_gpscripts
	Purpose : To initialize the maintenance and read/write scripts.  This
		  function is called by the core validation code and need not
		  be called by the test writer.
	Input: Nothing
	Output: On successful alignment and script initialization the function
		returns SCSI_GOOD.  If there are any memory allocation problems
		then the function returns SCSI_COMMANDFAILED.
	Assumptions: none
	Other functions called: malloc to allocate memory for ti tables 
				ByteAlignBuffer to align ti tables
				align_script to ensure script dword alignment
				sizeof to get size of script arrays
				getPhysAddr to initialize script buffers.

******************************************************************************/

VINT init_gpscripts(void)
{

	 VUBYTE far *buf_ptr;   /* temp ptr to ti tables */

	 /* allocate space for maint table */
	 buf_ptr =  (VUBYTE far *) malloc((MAINT_TABLE_SIZE * sizeof(ti_entry))+ 4);
	 /* did we get the memory */
	 if (buf_ptr == NULL) return(SCSI_COMMANDFAILED);

	 /* dword align the table buffer */
	 maint_buffer_table = (ti_entry far *) ByteAlignBuffer(buf_ptr, 0);

	 /* allocate space for rw table */
	 buf_ptr =  (VUBYTE far *) malloc((RW_TABLE_SIZE * sizeof(ti_entry))+ 4);
	 /* did we get the memory */
	 if (buf_ptr == NULL) return(SCSI_COMMANDFAILED);

	 /* dword align the table buffer */
	 rw_buffer_table = (ti_entry far *) ByteAlignBuffer(buf_ptr, 0);

	 /* align the scripts */
	 script = align_script(MAINT_SCRIPT, sizeof(MAINT_SCRIPT));
	 /* did we get the memory, if it was necessary to relocate the script */
	 if (script == NULL) return(SCSI_COMMANDFAILED);

	 rw_script = align_script(RW_SCRIPT, sizeof(RW_SCRIPT));
	 /* did we get the memory, if it was necessary to relocate the script */
	 if (script == NULL) return(SCSI_COMMANDFAILED);

	 /* init some externs used in the rw script */
	 rw_script[(VINT) E_RW_resel_msg_buf_Used[0]] = getPhysAddr(&msg_reject_buf);
	 rw_script[(VINT) E_RW_Bitbucket_buf_Used[0]] = getPhysAddr(&bitbucket);
	 rw_script[(VINT) E_RW_Bitbucket_buf_Used[1]] = getPhysAddr(&bitbucket);
	 rw_script[(VINT) E_RW_Bitbucket_buf_Used[2]] = getPhysAddr(&bitbucket);

	 /* init physical address variable used to find where we are in the */
	 /* Scatter/Gather list */
	 FirstDIMove_paddr = getPhysAddr(&rw_script[(VINT) Ent_RW_Offset_patch_di/4]) + 8;
	 FirstDOMove_paddr = getPhysAddr(&rw_script[(VINT) Ent_RW_Offset_patch_do/4]) + 8;

	 /* init the Rqsns_or_Inquiry variable that indicates to do a RS or an INQ */
	 /* default is Rqsns */
	 Rqsns_or_Inquiry = 0x03;

	 return(SCSI_GOOD);

}


/******************************************************************************
 Function:  init_scsi
	Purpose : Find and initialize all scsi devices on the bus optionally
		  negotiating sync and wide.
	Input:  Pointer to pcidev_record structure to indicate the PCI device
		whose scsi bus to initialize.
		Flag do_sync to indicate whether or not to do sync negotiation
		or not (0 don't do sync neg., 1 do sync neg.)  The function
		will negotiate for the fastest the PCI device can transfer data.
		Flag do_wide to indicate whether or not to do wide negotiation
		or not (0 don't do wide neg., 1 do wide neg.)
	Output: A pointer to a linked list of scsidev_record strutures.  Each
		entry in the list represents a device located on the scsi bus.
		All important information about the device is contained in this
		structure.  See 8xx_scsi.h for a detailed description of the
		scsidev_record structure.
	Assumptions:  none
	Restrictions: If a device hangs the bus during this function it will
					  no longer function properly.
	Other functions called: SCSI routines are called in this order:
									do_tur, do_rqsense, do_inquiry, do_rdcap
									and optionally do_wideneg and do_syncneg
									IORead8 to read chip registers
									malloc to get memory for scsi device entry
									sizeof to get size of scsi device record
	Notes: This function will attempt to clear all check conditions prior
			 to continuing to look at a device.  This includes clearing power on
			 reset or issuing a start unit to spin the device up.
			 This function will reset the SCSI bus.
			 This function will only do wide negotiation if requested, the drive
			 supports it and the chip supports it.
			 If the 8xx device being used is not recognized this function will
			 return NULL.
			 This function will not scan the device's scsi id, it will skip
			 that id.

******************************************************************************/

scsidev_record *init_scsi(pcidev_record *PCIDevice, VINT do_sync, VINT do_wide)

{

	/* variables */
	VUBYTE dev_id;                /* Used to loop through dev id's during scan */
	VUBYTE dev_num;               /* Used to count actual dev's found */
	VUBYTE max_scan;              /* Set to 7 or 15 depending on device type */
	VUBYTE error;                 /* Flag set when an error is detected during a scan */
	VUBYTE default_scntl3;        /* Holds default scntl3 value when scanning */
	VUBYTE chip_id;               /* Holds device's scsi id */
	VUBYTE retry_count;           /* Counts retries when trying to clear check cond */
	VCHAR id_string[25];          /* Used to hold id string from inquiry */
	VCHAR rev_string[5];          /* Used to hold rev string from inquiry */
	scsimaint maint;              /* Maintenance structure for calling SCSI routines */
	c8xxdev_record C8XXDevice;    /* Holds device specific info */

	/* main structure for holding scsi device info */
	scsidev_record *SCSIDevice;

	/* temporary structures for holding current device info */
	scsidev_record *current_dev;
	scsidev_record tempdev;

	/* initialize device table */
	SCSIDevice = NULL;

	/* initialize the device count counter */
	dev_num = 0;

	/* make sure this is a device we recognize */
	if (C8XX_GetChipType(PCIDevice, &C8XXDevice) != C8XX_GOOD)
		return(NULL);

	/* now reset the SCSI bus */
	IOWrite8(PCIDevice->base_addr2+SCNTL1, 0x08);
	delay(3);
	/* clear the SCSI reset */
	IOWrite8(PCIDevice->base_addr2+SCNTL1, 0x00);

	/* clear the interrupt from the SCSI reset, quickest to read the */
	/* entire dword */
	IORead32(PCIDevice->base_addr2+SIEN0);

	if (C8XXDevice.features & C8XX_FAST20)
		default_scntl3 = 0x05;
	else
		default_scntl3 = 0x03;

	if (C8XXDevice.features & C8XX_WIDE_SCSI)
		max_scan = 15;
	else
		max_scan = 7;

	/* get the chip's SCSI id for checking later */
	chip_id = IORead8(PCIDevice->base_addr2+SCID) & 0x0F;

	WriteLog(LOG_PASS, LOG_DETAIL, "Initializing SCSI Bus...\n", NULL);

	/* The following loop executes the following sequence of scsi commands */
	/* to a device on the scsi bus:                                        */
	/*  Test Unit Ready                                                    */
	/*  Request Sense                                                      */
	/*  Inquiry                                                            */
	/*  Read Capacity                                                      */
	/*  Wide Neqotiation (via TUR)      (Optional)                         */
	/*  Sync Negotiation (via TUR) (Optional)                              */
	/* If any of the above commands fails then the device at that id is    */
	/* ignored and the scan goes on to the next id                         */
	/* Note: Wide and Sync are only done if the drive supports it and the  */
	/* do_sync and do_wide flags are set                                                                                      */

	for (dev_id=0; dev_id<max_scan; dev_id++)
	{

		/* If we are at the chip's id then skip scanning this id */
		if (chip_id == dev_id) continue;

		/* initialize error flag */
		error = 0;

		/* initialize tempdev structure */
		tempdev.id_type = 0;
		tempdev.flags = 0;
		/* this is the default selection value, changed by sync and wide */
		/* code if necessary */
		tempdev.sel_info = ((VULONG) default_scntl3 << 24) |
			((VULONG) dev_id << 16);
		tempdev.maxlba = 0;
		tempdev.blksize = 0;
		tempdev.nxt_dev = NULL;

		/* initialize device id table info */
		maint.sel_info =  tempdev.sel_info;

		/* a single data buffer is used for init so we assign it here once */
		maint.data_buf = buffer;

		if (do_TUR(PCIDevice, &maint) == SCSI_GOOD)
		{

			/* init retry count */
			retry_count = 0;

			while ((maint.stat_buf == 2) && (retry_count <= INITSCSI_MAX_RETRIES)
						&& !error)
			{
				/* do requst sense */
				/* set up necessary variables */
				maint.data_length = 14;
				/* error during request sense */
				if (do_RqSense(PCIDevice, &maint) != SCSI_GOOD) error = 1;

				/* make sure they gave use all the data we wanted */
				if (maint.data_length < 14)
					/* if not we can't mess with this target */
					error = 1;
				else
				{
					/* handle sense data */
					if (maint.data_buf[12] == 0x04) /* LUN Not Ready */
					{
						switch (maint.data_buf[13])
						{
							case 0x00:                      /* not ready, unknown cause */
							{
								/* try and start the device */
								/* set the cmd_buf for start unit */
								maint.cmd_buf[4] = 0x01;
								if (do_StartStopUnit(PCIDevice, &maint) != SCSI_GOOD)
									error = 1;
								retry_count++;          /* try again */
								break;
							}
							case 0x01:                      /* not ready, becoming ready */
							{
								/* wait a while then retry */
								delay(100);
								retry_count++;
								break;
							}
							case 0x02:                      /* not ready, becoming ready */
							{
								/* try and start the device */
								/* set the cmd_buf for start unit */
								maint.cmd_buf[4] = 0x01;
								if (do_StartStopUnit(PCIDevice, &maint) != SCSI_GOOD)
									error = 1;
								retry_count++;          /* try again */
								break;
							}
							default:                                /* all other ASCQ's will error out */
							{
								error = 1;
								break;
							}
						}
					}
					else if (maint.data_buf[12] == 0x29)    /* power on/reset */
						/* just retry this one */
						retry_count++;
					else                                                                            /* error on everything else */
						error = 1;

					/* if no error now do a TUR to check for more check cond. */
					if (!error)
					{
						if (do_TUR(PCIDevice, &maint) != SCSI_GOOD)
							error = 1;
					}
				}
			}

			/* check if we hit max retries */
			if (retry_count > INITSCSI_MAX_RETRIES)
				/* if yes, error out */
				error = 1;

			/* do inquiry */
			if (!error)
			{
				/* set up data length variable */
				maint.data_length = 0x24;
				/* error during inquiry */
				if (do_Inquiry(PCIDevice, &maint) == SCSI_GOOD)
				{
					if ((((maint.data_buf[0] & 0xE0)>>5) <= 1)
						&& ((maint.data_buf[2] & 0x07) == 2))
					{                                                                                                                       /* make sure device is there */
							/* print device number, id and name etc. to screen */
							strncpy(id_string, &maint.data_buf[8], 24);
							id_string[24] = '\0';
							strncpy(rev_string, &maint.data_buf[32], 4);
							rev_string[4] = '\0';
							sprintf(string, "Device #%d: ID %d %s Rev:%s Wide:%d TCQ:%d Sync:%d\n",dev_num, dev_id,
								id_string, rev_string, (maint.data_buf[7]&0x20)>>5,
								(maint.data_buf[7]&0x02)>>1, (maint.data_buf[7]&0x10)>>4);
							WriteLog(LOG_PASS, LOG_DETAIL, string, NULL);

							/* update tempdev info */
							/* store id of device */
							tempdev.id_type = (dev_id<<4) | (maint.data_buf[0]&0x0f);

							/* we currently don't care about the ISO or ECMA version */
							/* we check to make sure device is scsi2 or we don't recognize it */
							/* we currently don't care about AENC, TermIOP, or Inquiry response */
							/* data format */

							/* check additional length to make sure data is valid */
							if (maint.data_buf[4] >= 3)     /* only want to make sure byte 8 */
							{                                                                                       /* of inquiry data is there */
								/* these flags are set to indicate if the device will support */
								/* the functions, not if we are actually utilizing them at the*/
								/* time.                                                            */
								tempdev.flags = ((maint.data_buf[1]&0x80)>>4) |  /* removable */
										((maint.data_buf[7]&0x02)<<1) |  /* cmd queue */
										((maint.data_buf[7]&0x20)>>4) |  /* wide */
										((maint.data_buf[7]&0x10)>>4);   /* sync */
							}
					}
					else error = 1;     /* we don't support any vendor specific */
											  /* qualifiers right now, or non SCSI2 dev */
				}
				else error = 1;

				if (!error)
				{
					if (do_RdCap(PCIDevice, &maint) == SCSI_GOOD)
					{
						/* update tempdev with info from read cap */
						tempdev.maxlba = ((VULONG) maint.data_buf[0] << 24) |
							((VULONG) maint.data_buf[1] << 16) |
							((VULONG) maint.data_buf[2] << 8)  |
							maint.data_buf[3];
						tempdev.blksize = ((VULONG) maint.data_buf[4] << 24) |
							((VULONG) maint.data_buf[5] << 16) |
							((VULONG) maint.data_buf[6] << 8)  |
							maint.data_buf[7];
					}
					else error = 1;
				}

				/* only do wide negotiation if the drive supports it the */
				/* flag do_wide is set and the chip supports it */
				if (!error && (tempdev.flags&0x02) && do_wide
					&& (C8XXDevice.features & C8XX_WIDE_SCSI))
				{
					/* set up wide message */
					maint.msgo_buf[4] = 0x01;

					do_WideNeg(PCIDevice, &maint);

					tempdev.sel_info = maint.sel_info;

				}

				/* only do sync negotiation if the drive supports it and the */
				/* flag do_sync is set */
				if (!error && (tempdev.flags&0x01) && do_sync)
				{
					/* set up sync values */
					if (C8XXDevice.features & C8XX_FAST20)
							maint.msgo_buf[4] = 0x0C;         /* negotiate for fast-20 */
					else maint.msgo_buf[4] = 0x19;          /* negotiate for 10 mb/s */

					/* negotiate for max offset 8xx device can do */
					maint.msgo_buf[5] = C8XXDevice.max_offset;

					do_SyncNeg(PCIDevice, &maint);

					/* copy the selection info over to the device structure */
					/* the value may not have changed if sync neg failed */
					tempdev.sel_info = maint.sel_info;
				}

				if (!error)
				{
					/* we found the device so create a new device entry */
					if (SCSIDevice == NULL)
					{                                                                                               /* first device in table */
						SCSIDevice = (scsidev_record *) malloc(sizeof(scsidev_record));
						SCSIDevice->nxt_dev = NULL;     /* terminate table */
						current_dev = SCSIDevice;       /* set the current device to be */
																  /* the first device */
					}
					else                             /* other devices in table */
					{                                        /* find the end */

						current_dev = SCSIDevice;               /* start at beginning */
						while (current_dev->nxt_dev != NULL)   /* find last entry */
							current_dev = current_dev->nxt_dev;

						current_dev->nxt_dev = (scsidev_record *) malloc(sizeof(scsidev_record));
						current_dev = current_dev->nxt_dev;    /* set pointer to newly */
											/* allocated entry */
						current_dev->nxt_dev = NULL;            /* terminate table */
					}
					/* now current dev points to the entry in the device table that we */
					/* want to modify */

					/* copy device info into device table */

					current_dev->id_type = tempdev.id_type;
					current_dev->flags = tempdev.flags;
					current_dev->sel_info = tempdev.sel_info;
					current_dev->maxlba = tempdev.maxlba;
					current_dev->blksize = tempdev.blksize;

					/* increment the device count */
					dev_num++;
				}
			}
		}
	}

	return(SCSIDevice);

}


/******************************************************************************
 Function:  do_tur
	Purpose :  To execute a Test Unit Ready to a device on the SCSI bus
	Input:  A pointer to the pcidev_record data structure and the scsimaint
			  data structure.  The following variables in the scsimaint structure
			  must be initialized for do_tur to function properly:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
			  All other variables in the structure need not be initialized.
	Output: Upon successful completion do_tur returns SCSI_GOOD.
			  Upon an ISTAT polling timeout (no interrupt) do_tur returns SCSI_POLLTO
			  Upon a selection timeout do_tur returns SCSI_SELECTTO
			  Upon any other type of interrupt do_tur returns SCSI_ILLEGAL_INT
			  Upon any other error condition do_tur returns SCSI_UNKNOWN
			  The following variables in the scsimaint data structure will be
			  modified:
					cmd_buf will contain the TUR command
					msgo_buf[0] will contain the identify message 0x80
					stat_buf will contain the status from the target
					msgi_buf will contain the final message in byte from the target
	Assumptions: none
	Restrictions: Disconnects are not allowed with this command.
	Other functions called:  IOWriteDSP to start the Script engine
									 IOWrite32 to initialize the DSA register
									 pollISTAT to look for an interrupt
									 getPhysAddr to patch TI entries
									 Fail_SCSI to clean up on an error condition

******************************************************************************/

VINT do_TUR(pcidev_record *PCIDevice, scsimaint *maint)

{
	VINT int_return;                        /* Holds return value from PollISTAT */
	VUBYTE dstat, istat;            /* Holds interrupt values from PollISTAT */
	VUINT sist;
	VULONG vector;

	/* setup the DSA to point to the Maint TI table */
	IOWrite32(PCIDevice->base_addr2+DSA, getPhysAddr(maint_buffer_table));

	/* setup the clock divider and target id info */
	maint_buffer_table[MAINT_SCSI_ID].count = maint->sel_info;
	maint_buffer_table[MAINT_SCSI_ID].address = 0x00000000;
	maint_buffer_table[MAINT_MSGO_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGO_BUF].address = getPhysAddr(maint->msgo_buf);
	maint_buffer_table[MAINT_CMD_BUF].count = 0x00000006;
	maint_buffer_table[MAINT_CMD_BUF].address = getPhysAddr(maint->cmd_buf);
	maint_buffer_table[MAINT_STAT_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_STAT_BUF].address = getPhysAddr(&maint->stat_buf);
	maint_buffer_table[MAINT_MSGI_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGI_BUF].address = getPhysAddr(&maint->msgi_buf);

	/* Initialize the message out command, caller can't set this */
	maint->msgo_buf[0] = 0x80;

	/* Initialize TUR command */
	maint->cmd_buf[0] = 0x00;
	maint->cmd_buf[1] = 0x00;
	maint->cmd_buf[2] = 0x00;
	maint->cmd_buf[3] = 0x00;
	maint->cmd_buf[4] = 0x00;
	maint->cmd_buf[5] = 0x00;

	/* start script */
	iowriteDSP(PCIDevice->base_addr2, getPhysAddr(script)
		+ Ent_maint_scsi);


	/* wait for interrupt to complete */
	int_return = PollISTAT(PCIDevice->base_addr2, &vector, &dstat,
		&istat, &sist, 2);

	if (int_return == C8XX_GOOD)
	{
		if ((dstat & C8XX_INT_SIR) && (vector == A_maint_cmd_complete))
			 return(SCSI_GOOD);

		else if (sist & C8XX_INT_STO)
			return(SCSI_SELECTTO);

		else
			return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_ILLEGAL_INT));
	}
	else if (int_return == C8XX_POLLTO)
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_POLLTO));

	else
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_UNKNOWN));
}


/******************************************************************************
 Function:  do_RqSense
	Purpose :  To execute a Request Sense or Inquiry to a device on the SCSI bus
	Input:  A pointer to the pcidev_record data structure and the scsimaint
			  data structure.  The following variables in the scsimaint structure
			  must be initialized for do_RqSense to function properly:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					data_length must be set to the desired amount of bytes to be
						recieved
					data_buf must be initialized to point to a valid data buffer
						this buffer must be at least 'data_length' in size
			  All other variables in the structure need not be initialized.
			  The Rqsns_or_Inquiry variable is set to the command byte of one
					or the other commands.
	Output: Upon successful completion do_rqsense returns SCSI_GOOD.
			  Upon an ISTAT polling timeout (no interrupt) do_rqsense returns SCSI_POLLTO
			  Upon a selection timeout do_rqsense returns SCSI_SELECTTO
			  Upon any other type of interrupt do_rqsense returns SCSI_ILLEGAL_INT
			  Upon any other error condition do_rqsense returns SCSI_UNKNOWN
			  The following variables in the scsimaint data structure will be
			  modified:
					cmd_buf will contain the Request Sense or Inquiry command
					msgo_buf[0] will contain the identify message 0x80
					data_buf will be filled with the Rq Snse or Inquiry data
					data_length will reflect the actual amount of data received
					stat_buf will contain the status from the target
					msgi_buf will contain the final message in byte from the target
	Assumptions: none
	Restrictions: Disconnects are not allowed with this command.
	Other functions called:  IOWriteDSP to start the Script engine
									 IOWrite32 to initialize the DSA register
									 pollISTAT to look for an interrupt
									 getPhysAddr to patch TI entries
									 Fail_SCSI to clean up on an error condition

******************************************************************************/

VINT do_RqSense(pcidev_record *PCIDevice, scsimaint *maint)

{

	VINT int_return;                        /* Holds return value from PollISTAT */
	VUBYTE dstat, istat;            /* Holds interrupt values from PollISTAT */
	VUINT sist;
	VULONG vector;

	/* setup the DSA to point to the Maint TI table */
	IOWrite32(PCIDevice->base_addr2+DSA, getPhysAddr(maint_buffer_table));

	/* fill in table indirect information */
	maint_buffer_table[MAINT_SCSI_ID].count = maint->sel_info;
	maint_buffer_table[MAINT_SCSI_ID].address = 0x00000000;
	maint_buffer_table[MAINT_MSGO_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGO_BUF].address = getPhysAddr(maint->msgo_buf);
	maint_buffer_table[MAINT_CMD_BUF].count = 0x00000006;
	maint_buffer_table[MAINT_CMD_BUF].address = getPhysAddr(maint->cmd_buf);
	maint_buffer_table[MAINT_STAT_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_STAT_BUF].address = getPhysAddr(&maint->stat_buf);
	maint_buffer_table[MAINT_MSGI_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGI_BUF].address = getPhysAddr(&maint->msgi_buf);
	maint_buffer_table[MAINT_DATA_BUF].count = (VUBYTE) maint->data_length;
	maint_buffer_table[MAINT_DATA_BUF].address = getPhysAddr(maint->data_buf);

	/* Initialize the message out command, caller can't set this */
	maint-> msgo_buf[0] = 0x80;

	/* Init RqSense command */
	maint->cmd_buf[0] = Rqsns_or_Inquiry;
	maint->cmd_buf[1] = 0x00;
	maint->cmd_buf[2] = 0x00;
	maint->cmd_buf[3] = 0x00;
	maint->cmd_buf[4] = (VUBYTE) maint->data_length;
	maint->cmd_buf[5] = 0x00;

	/* start script */
	iowriteDSP(PCIDevice->base_addr2, getPhysAddr(script)
		+ Ent_maint_scsi);

	while(1)
	{
		/* wait for interrupt to complete */
		int_return = PollISTAT( PCIDevice->base_addr2, &vector, &dstat,
			&istat, &sist, 2);

		if (int_return == C8XX_GOOD)
		{
			if ((dstat & C8XX_INT_SIR) && (vector == A_maint_cmd_complete))
				return(SCSI_GOOD);

			else if (sist & C8XX_INT_STO)
				return(SCSI_SELECTTO);

			else if (sist & C8XX_INT_MA)
			{
				/* there should be no data in the chip, it should be ok not to check */
				/* set data_length to the actual number of bytes received */
				maint->data_length -= IORead32(PCIDevice->base_addr2+DBC)
					& 0x00FFFFFFl;
				/* start up scripts again */
				RMWon(PCIDevice->base_addr2+DCNTL, 0x04);
				/* loop here, we should get a command complete int */
			}
			else
				return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_ILLEGAL_INT));
		}

		else if (int_return == C8XX_POLLTO)
			return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_POLLTO));

		else
			return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_UNKNOWN));
	}
}


/******************************************************************************
 Function:  do_Inquiry
	Purpose :  To execute an Inquiry to a device on the SCSI bus
	Input:  A pointer to the pcidev_record data structure and the scsimaint
			  data structure.  The following variables in the scsimaint structure
			  must be initialized for do_inquiry to function properly:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					data_length must be set to the desired amount of bytes to be
						recieved
					data_buf must be initialized to point to a valid data buffer
						this buffer must be at least 'data_length' in size
			  All other variables in the structure need not be initialized.
	Output: Upon successful completion do_inquiry returns SCSI_GOOD.
			  Upon an ISTAT polling timeout (no interrupt) do_inquiry returns SCSI_POLLTO
			  Upon a selection timeout do_inquiry returns SCSI_SELECTTO
			  Upon any other type of interrupt do_inquiry returns SCSI_ILLEGAL_INT
			  Upon any other error condition do_inquiry returns SCSI_UNKNOWN
			  The following variables in the scsimaint data structure will be
			  modified:
					cmd_buf will contain Inquiry command
					msgo_buf[0] will contain the identify message 0x80
					data_buf will be filled with the Inquiry data
					data_length will reflect the actual amount of data received
					stat_buf will contain the status from the target
					msgi_buf will contain the final message in byte from the target
	Assumptions: none
	Restrictions: Disconnects are not allowed with this command.
	Other functions called:  IOWriteDSP to start the Script engine
									 IOWrite32 to initialize the DSA register
									 pollISTAT to look for an interrupt
									 getPhysAddr to patch TI entries
									 Fail_SCSI to clean up on an error condition

******************************************************************************/

VINT do_Inquiry(pcidev_record *PCIDevice, scsimaint *maint)

{

	VINT ret;

	/* this just calls the do_Rqsense command with the global */
	/* Rqsns_or_Inquiry variable changed since the code is identicle */
	/* this saves code space but is invisible to the caller */
	Rqsns_or_Inquiry = 0x12;

	ret = do_RqSense(PCIDevice, maint);

	/* Set the flag back to Request sense commands */
	Rqsns_or_Inquiry = 0x03;

	return(ret);

}


/******************************************************************************
 Function:  do_RdCap
	Purpose :  To execute a Read Capicity to a device on the SCSI bus
	Input:  A pointer to the pcidev_record data structure and the scsimaint
			  data structure.  The following variables in the scsimaint structure
			  must be initialized for do_RdCap to function properly:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					data_buf must be initialized to point to a valid data buffer
						this buffer must be at least 8 bytes in size
			  All other variables in the structure need not be initialized.
	Output: Upon successful completion do_RdCap returns SCSI_GOOD.
			  Upon an ISTAT polling timeout (no interrupt) do_RdCap returns SCSI_POLLTO
			  Upon a selection timeout do_RdCap returns SCSI_SELECTTO
			  Upon any other type of interrupt do_RdCap returns SCSI_ILLEGAL_INT
			  Upon any other error condition do_RdCap returns SCSI_UNKNOWN
			  The following variables in the scsimaint data structure will be
			  modified:
					cmd_buf will contain the Read Capacity command
					msgo_buf[0] will contain the identify message 0x80
					data_buf will be filled with the 8 bytes of Read Capacity data
					stat_buf will contain the status from the target
					msgi_buf will contain the final message in byte from the target
	Assumptions: none
	Restrictions: Disconnects are not allowed with this command.
	Other functions called:  IOWriteDSP to start the Script engine
									 IOWrite32 to initialize the DSA register
									 pollISTAT to look for an interrupt
									 getPhysAddr to patch TI entries
									 Fail_SCSI to clean up on an error condition

******************************************************************************/

VINT do_RdCap(pcidev_record *PCIDevice, scsimaint *maint)

{

	VINT int_return;                        /* Holds return value from PollISTAT */
	VUBYTE dstat, istat;            /* Holds interrupt values from PollISTAT */
	VUINT sist;
	VULONG vector;

	/* setup the DSA to point to the Maint TI table */
	IOWrite32(PCIDevice->base_addr2+DSA, getPhysAddr(maint_buffer_table));

	maint_buffer_table[MAINT_SCSI_ID].count = maint->sel_info;
	maint_buffer_table[MAINT_SCSI_ID].address = 0x00000000;
	maint_buffer_table[MAINT_MSGO_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGO_BUF].address = getPhysAddr(maint->msgo_buf);
	maint_buffer_table[MAINT_CMD_BUF].count = 0x0000000A;
	maint_buffer_table[MAINT_CMD_BUF].address = getPhysAddr(maint->cmd_buf);
	maint_buffer_table[MAINT_STAT_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_STAT_BUF].address = getPhysAddr(&maint->stat_buf);
	maint_buffer_table[MAINT_MSGI_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGI_BUF].address = getPhysAddr(&maint->msgi_buf);
	maint_buffer_table[MAINT_DATA_BUF].count = 0x00000008;
	maint_buffer_table[MAINT_DATA_BUF].address = getPhysAddr(maint->data_buf);

	/* Initialize the message out command, caller can't set this */
	maint-> msgo_buf[0] = 0x80;

	/* Init RqSense command */
	maint->cmd_buf[0] = 0x25;
	maint->cmd_buf[1] = 0x00;
	maint->cmd_buf[2] = 0x00;
	maint->cmd_buf[3] = 0x00;
	maint->cmd_buf[4] = 0x00;
	maint->cmd_buf[5] = 0x00;
	maint->cmd_buf[6] = 0x00;
	maint->cmd_buf[7] = 0x00;
	maint->cmd_buf[8] = 0x00;
	maint->cmd_buf[9] = 0x00;

	/* start script */
	iowriteDSP(PCIDevice->base_addr2, getPhysAddr(script)
		+ Ent_maint_scsi);

	/* wait for interrupt to complete */
	int_return = PollISTAT( PCIDevice->base_addr2, &vector, &dstat,
		&istat, &sist, 2);

	if (int_return == C8XX_GOOD)
	{
		if ((dstat & C8XX_INT_SIR) && (vector == A_maint_cmd_complete))
			 return(SCSI_GOOD);

		else if (sist & C8XX_INT_STO)
			return(SCSI_SELECTTO);

		else
			return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_ILLEGAL_INT));
	}
	else if (int_return == C8XX_POLLTO)
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_POLLTO));

	else
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_UNKNOWN));
}

/******************************************************************************
 Function:  do_StartStopUnit
	Purpose :  To execute a Start or Stop Unit to a device on the SCSI bus
	Input:  A pointer to the pcidev_record data structure and the scsimaint
			  data structure.  The following variables in the scsimaint structure
			  must be initialized for do_tur to function properly:
					cmd_buf[4] must contain a valid start or stop command.
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
			  All other variables in the structure need not be initialized.
	Output: Upon successful completion do_StartStopUnit returns SCSI_GOOD.
			  Upon an ISTAT polling timeout (no interrupt) do_StartStopUnit
			  returns SCSI_POLLTO
			  Upon a selection timeout do_StartStopUnit returns SCSI_SELECTTO
			  Upon any other type of interrupt do_StartStopUnit returns SCSI_ILLEGAL_INT
			  Upon any other error condition do_StartStopUnit returns SCSI_UNKNOWN
			  The following variables in the scsimaint data structure will be
			  modified:
					cmd_buf will contain the Start or Stop unit command
					msgo_buf[0] will contain the identify message 0x80
					stat_buf will contain the status from the target
					msgi_buf will contain the final message in byte from the target
	Assumptions: none
	Restrictions: Disconnects are not allowed with this command.
	Other functions called:  IOWriteDSP to start the Script engine
									 IOWrite32 to initialize the DSA register
									 pollISTAT to look for an interrupt
									 getPhysAddr to patch TI entries
									 Fail_SCSI to clean up on an error condition

******************************************************************************/

VINT do_StartStopUnit(pcidev_record *PCIDevice, scsimaint *maint)

{
	VINT int_return;                        /* Holds return value from PollISTAT */
	VUBYTE dstat, istat;            /* Holds interrupt values from PollISTAT */
	VUINT sist;
	VULONG vector;

	/* setup the DSA to point to the Maint TI table */
	IOWrite32(PCIDevice->base_addr2+DSA, getPhysAddr(maint_buffer_table));

	/* setup the clock divider and target id info */
	maint_buffer_table[MAINT_SCSI_ID].count = maint->sel_info;
	maint_buffer_table[MAINT_SCSI_ID].address = 0x00000000;
	maint_buffer_table[MAINT_MSGO_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGO_BUF].address = getPhysAddr(maint->msgo_buf);
	maint_buffer_table[MAINT_CMD_BUF].count = 0x00000006;
	maint_buffer_table[MAINT_CMD_BUF].address = getPhysAddr(maint->cmd_buf);
	maint_buffer_table[MAINT_STAT_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_STAT_BUF].address = getPhysAddr(&maint->stat_buf);
	maint_buffer_table[MAINT_MSGI_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGI_BUF].address = getPhysAddr(&maint->msgi_buf);

	/* Initialize the message out command, caller can't set this */
	maint->msgo_buf[0] = 0x80;

	/* Initialize TUR command */
	maint->cmd_buf[0] = 0x1b;
	maint->cmd_buf[1] = 0x00;/* Set return status to when operation completes */
	maint->cmd_buf[2] = 0x00;
	maint->cmd_buf[3] = 0x00;
	/* cmd_buf[4] set by caller */
	maint->cmd_buf[5] = 0x00;

	/* start script */
	iowriteDSP(PCIDevice->base_addr2, getPhysAddr(script)
		+ Ent_maint_scsi);


	/* wait for interrupt to complete */
	/* The timeout for this command is set very high because it may */
	/* take quite a while for it to spin up or down */
	int_return = PollISTAT(PCIDevice->base_addr2, &vector, &dstat,
		&istat, &sist, 60);

	if (int_return == C8XX_GOOD)
	{
		if ((dstat & C8XX_INT_SIR) && (vector == A_maint_cmd_complete))
			 return(SCSI_GOOD);

		else if (sist & C8XX_INT_STO)
			return(SCSI_SELECTTO);

		else
			return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_ILLEGAL_INT));
	}
	else if (int_return == C8XX_POLLTO)
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_POLLTO));

	else
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_UNKNOWN));
}


/******************************************************************************
 Function:  do_SyncNeg
	Purpose :  To perform sync negotiation via a tur to a device on the SCSI bus
	Input:  A pointer to the pcidev_record data structure and the scsimaint
			  data structure.  The following variables in the scsimaint structure
			  must be initialized for do_SyncNeg to function properly:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					msgo_buf[4] must have the Sync Negotiation period as defined by
						the SCSI spec
					msgo_buf[5] must have the Sync Neqotiation offset as defined by
						the SCSI spec
			  All other variables in the structure need not be initialized.
	Output: Upon successful completion do_SyncNeg returns SCSI_GOOD.
			  Upon a Message Reject of the negotiation do_SyncNeg returns SCSI_MSGREJECT
			  Upon an ISTAT polling timeout (no interrupt) do_SyncNeg returns SCSI_POLLTO
			  Upon a selection timeout do_SyncNeg returns SCSI_SELECTTO
			  Upon any other type of interrupt do_SyncNeg returns SCSI_ILLEGAL_INT
			  Upon any other error condition do_SyncNeg returns SCSI_UNKNOWN
			  The following variables in the scsimaint data structure will be
			  modified:
					cmd_buf will contain the TUR command
					msgo_buf[0] will contain the identify message 0x80
					msgo_buf[1-3] will contain the sync neg message [0x01,0x03,0x01]
					emsgi_buf[0-3] will contain the sync neg reply from the target
					stat_buf will contain the status from the target
					msgi_buf will contain the final message in byte from the target
					sel_info SCNTL3 an SXFER values will be modified to meet the
						clock divider requirement of the negotiated sync offset and
						transfer rate
	Assumptions: none
	Restrictions: Disconnects are not allowed with this command.
					  Remember negotiation ordering should always be Wide then Sync
					  as Wide negotiation resets Sync values to Async.
	Other functions called:  IOWriteDSP to start the Script engine
									 IOWrite32 to initialize the DSA register
									 pollISTAT to look for an interrupt
									 getPhysAddr to patch TI entries
									 Fail_SCSI to clean up on an error condition

******************************************************************************/

VINT do_SyncNeg(pcidev_record *PCIDevice, scsimaint *maint)

{
	VINT int_return;                        /* Holds return value from PollISTAT */
	VUBYTE dstat, istat;            /* Holds interrupt values from PollISTAT */
	VUINT sist;
	VULONG vector;
	/* variables used to assigned sync reg values */
	VUBYTE SyncPeriod, SyncOffset, fast20En;
	VUBYTE WideState, scntl3, sxfer;
	VUBYTE *ptrTable;

	/* setup the DSA to point to the Maint TI table */
	IOWrite32(PCIDevice->base_addr2+DSA, getPhysAddr(maint_buffer_table));

	/* setup the clock divider and target id info */
	maint_buffer_table[MAINT_SCSI_ID].count = maint->sel_info;
	maint_buffer_table[MAINT_SCSI_ID].address = 0x00000000;
	maint_buffer_table[MAINT_MSGO_BUF].count = 0x00000006;
	maint_buffer_table[MAINT_MSGO_BUF].address = getPhysAddr(maint->msgo_buf);
	maint_buffer_table[MAINT_CMD_BUF].count = 0x00000006;
	maint_buffer_table[MAINT_CMD_BUF].address = getPhysAddr(maint->cmd_buf);
	maint_buffer_table[MAINT_STAT_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_STAT_BUF].address = getPhysAddr(&maint->stat_buf);
	maint_buffer_table[MAINT_MSGI_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGI_BUF].address = getPhysAddr(&maint->msgi_buf);
	maint_buffer_table[MAINT_EMSGI_BUF].count = 0x00000004;
	maint_buffer_table[MAINT_EMSGI_BUF].address = getPhysAddr(maint->emsgi_buf);

	/* Initialize the message out buffer, caller can't set this */
	maint->msgo_buf[0] = 0x80;
	maint->msgo_buf[1] = 0x01;    /* extended message */
	maint->msgo_buf[2] = 0x03;    /* 3 bytes to follow */
	maint->msgo_buf[3] = 0x01;    /* sync negotiation */
	/* caller sets up period and offset */

	/* Initialize TUR command */
	maint->cmd_buf[0] = 0x00;
	maint->cmd_buf[1] = 0x00;
	maint->cmd_buf[2] = 0x00;
	maint->cmd_buf[3] = 0x00;
	maint->cmd_buf[4] = 0x00;
	maint->cmd_buf[5] = 0x00;

	/* start script */
	iowriteDSP(PCIDevice->base_addr2, getPhysAddr(script)
		+ Ent_maint_scsi);


	/* wait for interrupt to complete */
	int_return = PollISTAT( PCIDevice->base_addr2, &vector, &dstat,
		&istat, &sist, 2);

	if (int_return == C8XX_GOOD)
	{
		if ((dstat & C8XX_INT_SIR) && (vector == A_maint_cmd_complete))
		{
			if (maint->msgi_buf == 0x07)                  /* got rejected */
				return(SCSI_MSGREJECT);

			else                                    /* sync neg actually worked */
			{
				/* get negotiated sync per and offset values from the buffer */
				SyncPeriod = maint->emsgi_buf[2];
				SyncOffset = maint->emsgi_buf[3];

				/* store the state of the wide bit */
				WideState = (VUBYTE) ((maint->sel_info & 0x08000000l) >> 24);

				fast20En = 0;                     /* Assume no fast 20 */
				ptrTable = &Sync40Table[ 0 ][ 0 ];

				/* if the compare below is true we assume an 80 MHz clock */
				if ((IORead8(PCIDevice->base_addr2+SCNTL3) & 0x07) == 0x5)
				{
					/*  If the rate is faster than 10 mega transfers per second, then
					 *  the table guarantees we will not bump the entry to 10 mega
					 *  transfers, so we must set the fast 20 flag.
					 */

					ptrTable = &Sync80Table[ 0 ][ 0 ];
					if (SyncPeriod < 25)
						fast20En = 0x80;
				}

				/*  Look through the table period entry, testing for the best match.
				 *  The match value becomes the period that is passed back to the
				 *  caller.
				 *
				 *  If the period given by the caller is faster than 10 mega transfers
				 *  and our chip doesn't handle this fact (ie. 40 MHz clock), then the
				 *  table entries already have handled setting the Scntl and Sxfer
				 *  values properly so that we run at 10 mega transfers.
				 *
				 *  NOTE:  We are guaranteed to find an entry in this table due to the
				 *         table values ranging from 0..255 and the SyncPeriod being
				 *         only an 8 bit value.
				 */

				while (SyncPeriod > ptrTable[ 0 ])
					ptrTable += 3;                  /* Skip one table column */

				scntl3 = fast20En | WideState | ptrTable[1];
				sxfer = ptrTable[2] | SyncOffset;

				/* clear current sync reg values in table indirect entry */
				maint->sel_info &= 0x00FF0000l;
				/* now put in new ones */
				maint->sel_info |= (((VULONG) scntl3 << 24) |
										 ((VULONG) sxfer << 8));

				return(SCSI_GOOD);
			}
		}

		else if (sist & C8XX_INT_STO)
			return(SCSI_SELECTTO);

		/* some drives just disconnect after sending a msg reject, the code */
		/* below handles this */
		else if ((sist & C8XX_INT_UDC) && (maint->msgi_buf == 0x07))
			return(SCSI_MSGREJECT);

		else
			return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_ILLEGAL_INT));
	}

	else if (int_return == C8XX_POLLTO)
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_POLLTO));

	else
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_UNKNOWN));
}


/******************************************************************************
 Function:  do_WideNeg
	Purpose :  To perform wide negotiation via a tur to a device on the SCSI bus
	Input:  A pointer to the pcidev_record data structure and the scsimaint
			  data structure.  The following variables in the scsimaint structure
			  must be initialized for do_WideNeg to function properly:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					msgo_buf[4] must have the Wide Negotiation value as defined by
						the SCSI spec (0=8-bit, 1=16-bit, 2=32-bit)
			  All other variables in the structure need not be initialized.
	Output: Upon successful completion do_WideNeg returns SCSI_GOOD.
			  Upon a Message Reject of the negotiation do_WideNeg returns SCSI_MSGREJECT
			  Upon an ISTAT polling timeout (no interrupt) do_WideNeg returns SCSI_POLLTO
			  Upon a selection timeout do_WideNeg returns SCSI_SELECTTO
			  Upon any other type of interrupt do_WideNeg returns SCSI_ILLEGAL_INT
			  Upon any other error condition do_WideNeg returns SCSI_UNKNOWN
			  The following variables in the scsimaint data structure will be
			  modified:
					cmd_buf will contain the TUR command
					msgo_buf[0] will contain the identify message 0x80
					msgo_buf[1-3] will contain the wide neg message [0x01,0x02,0x03]
					emsgi_buf[0-2] will contain the wide neg reply from the target
					stat_buf will contain the status from the target
					msgi_buf will contain the final message in byte from the target
					sel_info SCNTL3 values will be modified to so that the EWS bit
						is set appropriately.  SXFER values will be reset to 0x00
		  so that the chip will be reset to async as required in SCSI.

	Assumptions: none
	Restrictions: Disconnects are not allowed with this command.
					  Remember negotiation ordering should always be Wide then Sync
					  as Wide negotiation resets Sync values to Async.
					  do_wide will reset sync values (in SCNTL3 and SXFER) as
					  required in the scsi spec.
	Other functions called:  IOWriteDSP to start the Script engine
									 IOWrite32 to initialize the DSA register
									 pollISTAT to look for an interrupt
									 getPhysAddr to patch TI entries
									 Fail_SCSI to clean up on an error condition

******************************************************************************/

VINT do_WideNeg(pcidev_record *PCIDevice, scsimaint *maint)

{
	VINT int_return;                        /* Holds return value from PollISTAT */
	VUBYTE dstat, istat;            /* Holds interrupt values from PollISTAT */
	VUINT sist;
	VULONG vector;

	/* setup the DSA to point to the Maint TI table */
	IOWrite32(PCIDevice->base_addr2+DSA, getPhysAddr(maint_buffer_table));

	/* setup the clock divider and target id info */
	maint_buffer_table[MAINT_SCSI_ID].count = maint->sel_info;
	maint_buffer_table[MAINT_SCSI_ID].address = 0x00000000;
	maint_buffer_table[MAINT_MSGO_BUF].count = 0x00000005;
	maint_buffer_table[MAINT_MSGO_BUF].address = getPhysAddr(maint->msgo_buf);
	maint_buffer_table[MAINT_CMD_BUF].count = 0x00000006;
	maint_buffer_table[MAINT_CMD_BUF].address = getPhysAddr(maint->cmd_buf);
	maint_buffer_table[MAINT_STAT_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_STAT_BUF].address = getPhysAddr(&maint->stat_buf);
	maint_buffer_table[MAINT_MSGI_BUF].count = 0x00000001;
	maint_buffer_table[MAINT_MSGI_BUF].address = getPhysAddr(&maint->msgi_buf);
	maint_buffer_table[MAINT_EMSGI_BUF].count = 0x00000003;
	maint_buffer_table[MAINT_EMSGI_BUF].address = getPhysAddr(maint->emsgi_buf);

	/* Initialize the WDTR message out caller can't set this */
	maint->msgo_buf[0] = 0x80;
	maint->msgo_buf[1] = 0x01;
	maint->msgo_buf[2] = 0x02;
	maint->msgo_buf[3] = 0x03;
	/* caller sets the actual message (msgo_buf[4]) */

	/* Initialize TUR command */
	maint->cmd_buf[0] = 0x00;
	maint->cmd_buf[1] = 0x00;
	maint->cmd_buf[2] = 0x00;
	maint->cmd_buf[3] = 0x00;
	maint->cmd_buf[4] = 0x00;
	maint->cmd_buf[5] = 0x00;

	/* start script */
	iowriteDSP(PCIDevice->base_addr2, getPhysAddr(script)
		+ Ent_maint_scsi);


	/* wait for interrupt to complete */
	int_return = PollISTAT( PCIDevice->base_addr2, &vector, &dstat,
		&istat, &sist, 2);

	if (int_return == C8XX_GOOD)
	{
		if ((dstat & C8XX_INT_SIR) && (vector == A_maint_cmd_complete))
		{
			 if (maint->msgi_buf == 0x07)                  /* got rejected */
				return(SCSI_MSGREJECT);

			 else
			 {
				/* we successfully negotiated, so we must shut off sync */
				/* we do this by setting the sync offset to zero in sxfer */
				maint->sel_info &= 0xFFFFE000l;
				if (maint->emsgi_buf[2])                  /* negotiated for wide */
					maint->sel_info |= 0x08000000l;         /* set EWS bit in scntl3 */
				else
					maint->sel_info &= 0xF7FFFFFFl;         /* clear EWS bit in scntl3 */

				return(SCSI_GOOD);
			 }
		}
		else if (sist & C8XX_INT_STO)
			return(SCSI_SELECTTO);

		/* some drives just disconnect after sending a msg reject, the code */
		/* below handles this */
		else if ((sist & C8XX_INT_UDC) && (maint->msgi_buf == 0x07))
			return(SCSI_MSGREJECT);

		else
			return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_ILLEGAL_INT));
	}

	else if (int_return == C8XX_POLLTO)
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_POLLTO));

	else
		return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_UNKNOWN));
}


/******************************************************************************
 Function: do_Read6
	Purpose : To execute a 6 byte read to a SCSI device
	Input: A pointer to the pcidev_record structure and a pointer to the
			 scsirw data structure.  The follow variables in the scsirw
			 structure must be initialized:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					lba must be initialized to a valid target logical block address
					blk_count must have a valid block count for the read
					id_count must be 1 or 3 depending on whether or not tagged
						command queuing is used (1 for no TCQ 3 for TCQ)
					id_buf[0-2] must be initilized with a valid identify message and
						optionally a 2 byte TCQ message
					sg_count must be initialized to the number of scatter gather entries
						NOTE: If scatter gather is not being used this variable should
						be set to 1.
					sg_entry should point to a valid scatter gather list in table
						indirect format.  NOTE: If scatter gather is not being used
						this pointer should point to the single table indirect
						entry for the io.  The scatter gather list pointed to by
						sg_entry should match sg_count in length exactly.
			  All other variables in the structure need not be initialized.


	Output:  do_Read6 returns the return value from the do_rw function.
			please see the do_rw function for a detailed description of
			return values.
		 do_Read6 modifies the following variables in the scsirw data structure:
			cmd_buf is set to the appropriate read command given the
				settings of lba and blk_count.
		 Other variables are modified by the do_rw function.  Please see that
			function description for a listing of other variables modified.
	Assumptions:  None
	Restrictions: None
	Other functions called:  do_rw to actually execute the read io to the
					target.

******************************************************************************/

VINT do_Read6(pcidev_record *PCIDevice, scsirw *rw)

{

	/* Modify Read6 specific varibles */
	rw_buffer_table[RW_CMD_BUF].count = 0x00000006;

	/* set up the command */
	rw->cmd_buf[0] = 0x08;
	rw->cmd_buf[1] = (VUBYTE) ((rw->lba >> 16) & 0x1F);
	rw->cmd_buf[2] = (VUBYTE) ((rw->lba >> 8) & 0xFF);
	rw->cmd_buf[3] = (VUBYTE) rw->lba;
	rw->cmd_buf[4] = (VUBYTE) rw->blk_count;
	rw->cmd_buf[5] = 0x00;

	return(do_rw(PCIDevice, rw));

}


/******************************************************************************
 Function: do_Read10
	Purpose : To execute a 10 byte read to a SCSI device
	Input: A pointer to the pcidev_record structure and a pointer to the
			 scsirw data structure.  The follow variables in the scsirw
			 structure must be initialized:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					lba must be initialized to a valid target logical block address
					blk_count must have a valid block count for the read
					id_count must be 1 or 3 depending on whether or not tagged
						command queuing is used (1 for no TCQ 3 for TCQ)
					id_buf[0-2] must be initilized with a valid identify message and
						optionally a 2 byte TCQ message
					sg_count must be initialized to the number of scatter gather entries
						NOTE: If scatter gather is not being used this variable should
						be set to 1.
					sg_entry should point to a valid scatter gather list in table
						indirect format.  NOTE: If scatter gather is not being used
						this pointer should point to the single table indirect
						entry for the io.  The scatter gather list pointed to by
						sg_entry should match sg_count in length exactly.
			  All other variables in the structure need not be initialized.


	Output:  do_Read10 returns the return value from the do_rw function.
			please see the do_rw function for a detailed description of
			return values.
		 do_Read10 modifies the following variables in the scsirw data structure:
			cmd_buf is set to the appropriate read command given the
				settings of lba and blk_count.
		 Other variables are modified by the do_rw function.  Please see that
			function description for a listing of other variables modified.
	Assumptions:  None
	Restrictions: None
	Other functions called:  do_rw to actually execute the read io to the 
					target. 

******************************************************************************/

VINT do_Read10(pcidev_record *PCIDevice, scsirw *rw)

{

	/* Modify Read10 specific varibles */
	rw_buffer_table[RW_CMD_BUF].count = 0x0000000A;

	/* set up the command */
	rw->cmd_buf[0] = 0x28;
	rw->cmd_buf[1] = 0x00;
	rw->cmd_buf[2] = (VUBYTE) ((rw->lba >> 24) & 0xFF);
	rw->cmd_buf[3] = (VUBYTE) ((rw->lba >> 16) & 0xFF);
	rw->cmd_buf[4] = (VUBYTE) ((rw->lba >> 8) & 0xFF);
	rw->cmd_buf[5] = (VUBYTE) rw->lba;
	rw->cmd_buf[6] = 0x00;
	rw->cmd_buf[7] = (VUBYTE) (rw->blk_count >> 8);
	rw->cmd_buf[8] = (VUBYTE) rw->blk_count;
	rw->cmd_buf[9] = 0x00;

	return(do_rw(PCIDevice, rw));

}


/******************************************************************************
 Function: do_Write6                  
	Purpose : To execute a 6 byte write to a SCSI device
	Input: A pointer to the pcidev_record structure and a pointer to the
	       scsirw data structure.  The follow variables in the scsirw 
	       structure must be initialized:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					lba must be initialized to a valid target logical block address
					blk_count must have a valid block count for the write
					id_count must be 1 or 3 depending on whether or not tagged
						command queuing is used (1 for no TCQ 3 for TCQ)
					id_buf[0-2] must be initilized with a valid identify message and
						optionally a 2 byte TCQ message
					sg_count must be initialized to the number of scatter gather entries
						NOTE: If scatter gather is not being used this variable should
						be set to 1.
					sg_entry should point to a valid scatter gather list in table 
						indirect format.  NOTE: If scatter gather is not being used
						this pointer should point to the single table indirect
						entry for the io.  The scatter gather list pointed to by
						sg_entry should match sg_count in length exactly.
			  All other variables in the structure need not be initialized.


	Output:  do_Write6 returns the return value from the do_rw function.
			please see the do_rw function for a detailed description of
			return values.
		 do_Write6 modifies the following variables in the scsirw data structure:
			cmd_buf is set to the appropriate write command given the
				settings of lba and blk_count.
		 Other variables are modified by the do_rw function.  Please see that
			function description for a listing of other variables modified.
	Assumptions:  None
	Restrictions: None
	Other functions called:  do_rw to actually execute the write io to the 
					target. 

******************************************************************************/

VINT do_Write6(pcidev_record *PCIDevice, scsirw *rw)

{

	/* Modify Read6 specific varibles */
	rw_buffer_table[RW_CMD_BUF].count = 0x00000006;

	/* set up the command */
	rw->cmd_buf[0] = 0x0A;
	rw->cmd_buf[1] = (VUBYTE) ((rw->lba >> 16) & 0x1F);
	rw->cmd_buf[2] = (VUBYTE) ((rw->lba >> 8) & 0xFF);
	rw->cmd_buf[3] = (VUBYTE) rw->lba;
	rw->cmd_buf[4] = (VUBYTE) rw->blk_count;
	rw->cmd_buf[5] = 0x00;

	return(do_rw(PCIDevice, rw));

}


/******************************************************************************
 Function: do_Write10                  
	Purpose : To execute a 10 byte write to a SCSI device
	Input: A pointer to the pcidev_record structure and a pointer to the
	       scsirw data structure.  The follow variables in the scsirw 
	       structure must be initialized:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					lba must be initialized to a valid target logical block address
					blk_count must have a valid block count for the write
					id_count must be 1 or 3 depending on whether or not tagged
						command queuing is used (1 for no TCQ 3 for TCQ)
					id_buf[0-2] must be initilized with a valid identify message and
						optionally a 2 byte TCQ message
					sg_count must be initialized to the number of scatter gather entries
						NOTE: If scatter gather is not being used this variable should
						be set to 1.
					sg_entry should point to a valid scatter gather list in table 
						indirect format.  NOTE: If scatter gather is not being used
						this pointer should point to the single table indirect
						entry for the io.  The scatter gather list pointed to by
						sg_entry should match sg_count in length exactly.
			  All other variables in the structure need not be initialized.


	Output:  do_Write10 returns the return value from the do_rw function.
			please see the do_rw function for a detailed description of
			return values.
		 do_Write10 modifies the following variables in the scsirw data structure:
			cmd_buf is set to the appropriate write command given the
				settings of lba and blk_count.
		 Other variables are modified by the do_rw function.  Please see that
			function description for a listing of other variables modified.
	Assumptions:  None
	Restrictions: None
	Other functions called:  do_rw to actually execute the write io to the 
					target. 

******************************************************************************/

VINT do_Write10(pcidev_record *PCIDevice, scsirw *rw)

{

	/* Modify Read6 specific varibles */
	rw_buffer_table[RW_CMD_BUF].count = 0x0000000A;

	/* set up the command */
	rw->cmd_buf[0] = 0x2A;
	rw->cmd_buf[1] = 0x00;
	rw->cmd_buf[2] = (VUBYTE) ((rw->lba >> 24) & 0xFF);
	rw->cmd_buf[3] = (VUBYTE) ((rw->lba >> 16) & 0xFF);
	rw->cmd_buf[4] = (VUBYTE) ((rw->lba >> 8) & 0xFF);
	rw->cmd_buf[5] = (VUBYTE) rw->lba;
	rw->cmd_buf[6] = 0x00;
	rw->cmd_buf[7] = (VUBYTE) (rw->blk_count >> 8);
	rw->cmd_buf[8] = (VUBYTE) rw->blk_count;
	rw->cmd_buf[9] = 0x00;

	return(do_rw(PCIDevice, rw));

}


/******************************************************************************
 Function: do_rw
	Purpose : To execute a read or write type command to a SCSI device
	Input: A pointer to the pcidev_record structure and a pointer to the
			 scsirw data structure.  The follow variables in the scsirw
			 structure must be initialized:
					sel_info must have a valid table indirect selection entry
						remember ordering of the dword is (SCNTL3|SDID|SXFER|00)
					id_count must be 1 or 3 depending on whether or not tagged
						command queuing is used (1 for no TCQ 3 for TCQ)
					id_buf[0-2] must be initilized with a valid identify message and
						optionally a 2 byte TCQ message
					sg_count must be initialized to the number of scatter gather entries
						NOTE: If scatter gather is not being used this variable should
						be set to 1.
					sg_entry should point to a valid scatter gather list in table
						indirect format.  NOTE: If scatter gather is not being used
						this pointer should point to the single table indirect
						entry for the io.  The scatter gather list pointed to by
						sg_entry should match sg_count in length exactly.
					cmd_buf must be initialized with a valid read or write type command
					enable_auto_rqs must be initialized to either 0 to disable auto
						request sense or 1 (non-zero) to enable auto req. sense.
			  All other variables in the structure need not be initialized.
	Output: Upon successful completion do_rw returns SCSI_GOOD.
			  Upon an ISTAT polling timeout (no interrupt) do_rw returns SCSI_POLLTO
			  Upon a selection timeout do_rw returns SCSI_SELECTTO
			  Upon a phase mismatch during Status, Command or Message In
				phase do_rw returns SCSI_COMMANDFAILED or SCSI_CON_COMMANDFAILED
				depending on the state of the scsi bus after trying to clear it
				up with the bitbucket routine in the Fail_SCSI function
			  Upon any SCSI interrupts other that Phase Mismatch or Select TO
				do_rw returns SCSI_COMMANDFAILED or SCSI_CON_COMMANDFAILED
			  Upon any DMA interrupt other than SIR do_rw will return
				SCSI_COMMANDFAILED or SCSI_CON_COMMANDFAILED
			  Upon any DMA SIR interrupt other than Command_Complete, Need_More_SG
				or SWIDE_byte_moved vectors do_rw will return SCSI_COMMANDFAILED
				or SCSI_CON_COMMANDFAILED
			  Upon a SCSI CHECK CONDITION status do_rw will return
				SCSI_COMMANDFAILED or SCSI_CON_COMMANDFAILED
			  The following variables in the scsirw data structure are modified by
				the do_rw function:
					The data buffers pointed to by the sg_entry TI entries
						will be filled with the data from the target on
						a read operation.
					stat_buf will contain the status from the target
					msgi_buf will contain the final message in byte from the target
					resel_msg_buf will contain the final reselection identify and
						TCQ message (if used) if disconnects were allowed on
						this IO.
					If auto request sense is enabled and a SCSI CHECK CONDITION
						occurs during the IO then the sense_length will contain
						the number of valid sense bytes and sense_buffer will
						contain the sense data from the request sense.
					Three other Status values, IO_Major_Status, IO_Minor_Status and
					 Xfer_status will be filled in with various status values.
					The values are #defined in 8XX_SCSI.H. The following is a
					 description on how those statuses will be filled in:
					Xfer_status will be a 1 if all data was transferred or a 0 if
					 all data was not transferred.  This is independant of whether
					 or not the IO completed succesfully or not.
					Upon an Illegal SCSI interrupt:
						IO_Major_Status will contain SCSI_ILLEGAL_SCSI_INTERRUPT
						IO_Minor_Status will indicate the specific interrupt
					Upon an Illegal DMA interrupt:
						IO_Major_Status will contain SCSI_ILLEGAL_DMA_INTERRUPT
						IO_Minor_Status will indicate the specific interrupt
					Upon an Illegal SCRIPT interrupt:
						IO_Major_Status will contain SCSI_ILLEGAL_SCRIPT_INTERRUPT
						IO_Minor_Status will indicate the specific interrupt
					Upon a succesful SCSI read or write:
						IO_Major_Status will contain SCSI_COMMAND_COMPLETE
						IO_Minor_Status will contain SCSI_NO_STATUS
					Upon a SCSI check condition status:
						IO_Major_Status will contain SCSI_CHECK_CONDITION
						IO_Minor_Status will indicate whether or not the request
						 sense data is valid.  If enable_auto_rqs is 0 then this
						 will always indicate invalid sense data.
					All of the conditions below will only have the IO_Major_Status
					 set.  IO_Minor_Status will be set to SCSI_NO_STATUS.
						Upon a interrupt polling time out:
							IO_Major_Status will contain SCSI_CHECK_CONDITION
						Upon a SCSI Select Time Out:
							IO_Major_Status will contain SCSI_SELECTTO
	Assumptions:  That the byte counts pointed to be sg_entry add up to the
				blk_count * the Block Size (i.e. the number of bytes requested
				from the target = the number of bytes to be moved by the part)
				If this is not the case then the IO will die a miserable death
	Restrictions: This function is designed to do read or write type io's to
			SCSI block devices only.  The author does not gaurantee it's
			functionality if any other type of scsi command is used.
	Other functions called: IORead8/32 and IOWrite32 to Read and Write chip registers
				IOWriteDSP to start Scripts engine
				getPhysAddr to patch TI entries for IO
				HandleDataInPM to clean up after a Data In phase mismatch
				HandleDataOutPM to clean up after a Data Out phase mismatch
				Fill_SG_Entries to set up scatter gather entries for
					the io
				PollISTAT to look for interrupts during the IO
				Fail_SCSI to clean up on any failing condition.

******************************************************************************/

static VINT do_rw(pcidev_record *PCIDevice, scsirw *rw)

{

	VINT int_return;              /* Holds return value from PollISTAT */
	VUBYTE dstat, istat;          /* Holds interrupt values from PollISTAT */
	VUINT sist;
	VULONG vector;
	VINT NextSGEntry;             /* Holds next Scatter gather entry to be serviced */
	VINT scsi_command_complete;   /* Set when IO is complete to break loop */
	VINT Current_SG_Entry;        /* Holds TI entry being serviced when PM occurs */
	scsimaint maint;              /* for doing request sense */
	VINT sense_ret;               /* Return value from doing request sense */

	/* Set the DSA register */
	IOWrite32(PCIDevice->base_addr2+DSA, getPhysAddr(rw_buffer_table));

	rw_buffer_table[RW_SCSI_ID].count = rw->sel_info;
	rw_buffer_table[RW_SCSI_ID].address = 0x00000000;
	rw_buffer_table[RW_ID_BUF].count = rw->id_count;
	rw_buffer_table[RW_ID_BUF].address = getPhysAddr(rw->id_buf);
	rw_buffer_table[RW_CMD_BUF].address = getPhysAddr(rw->cmd_buf);
	rw_buffer_table[RW_STAT_BUF].count = 0x00000001;
	rw_buffer_table[RW_STAT_BUF].address = getPhysAddr(&rw->stat_buf);
	rw_buffer_table[RW_MSGI_BUF].count = 0x00000001;
	rw_buffer_table[RW_MSGI_BUF].address = getPhysAddr(&rw->msgi_buf);

	/* do the initial set up of the Scatter Gather entries */
	NextSGEntry = 0;
	Fill_SG_Entries(PCIDevice, rw, &NextSGEntry);

	if (rw->id_buf[0] & 0x40)               /* disconnects allowed */
	{       /* patch reselect buffer address info */
		rw_script[(VINT) E_RW_resel_msg_buf_Used[0]] = getPhysAddr(&rw->resel_msg_buf[0]);
		rw_script[(VINT) E_RW_resel_tag_buf1_Used[0]] = getPhysAddr(&rw->resel_msg_buf[1]);
		rw_script[(VINT) E_RW_resel_tag_buf2_Used[0]] = getPhysAddr(&rw->resel_msg_buf[2]);
	}

	/* clear the flag that indicates that all the data was transferred */
	RMWoff(PCIDevice->base_addr2+SBR, (VUBYTE) A_RW_ALL_DATA_TRANSFERRED);

	/* OK, now we're ready to start the script */
	iowriteDSP(PCIDevice->base_addr2, getPhysAddr(rw_script) + Ent_RW_Start);

	scsi_command_complete = 0;

	while (!scsi_command_complete)  /* loop until the command is done */
	{
		int_return = PollISTAT( PCIDevice->base_addr2, &vector, &dstat,
							&istat, &sist, 5);

		if (int_return == C8XX_POLLTO)
		{
			/* Set the IO status */
			rw->IO_major_status = SCSI_NO_INTERRUPT;
			rw->IO_minor_status = SCSI_NO_STATUS;
			/* Check for a complete transfer, unlikely if here */
			if (IORead8(PCIDevice->base_addr2+SBR) & A_RW_ALL_DATA_TRANSFERRED)
				rw->Xfer_status = SCSI_TRANSFER_COMPLETE;
			else
				rw->Xfer_status = SCSI_INCOMPLETE_TRANSFER;

			return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_POLLTO));
		}

		if (istat & C8XX_INT_SIP)               /* scsi interrupt */
		{
			/* the only acceptable interrupts here are Phase Mismatch and Sel TO */
			if (sist & C8XX_INT_STO)        /* sel TO, just return sel TO status */
			{
				/* Set the IO status */
				rw->IO_major_status = SCSI_SELECTTO;
				rw->IO_minor_status = SCSI_NO_STATUS;
				/* We know that the transfer didn't complete so just set that status */
				rw->Xfer_status = SCSI_INCOMPLETE_TRANSFER;

				return(SCSI_SELECTTO);
			}
			else if (sist & C8XX_INT_MA)    /* phase mismatch, could be good or bad */
			{
				/* depending on the phase we were in when we got the PM we */
				/* can handle it or need to get the bus quite and return an */
				/* error */
				switch (IORead8(PCIDevice->base_addr2+SOCL) & 0x07)
				{
					/* If we're in Data In or Data Out phase we need to do */
					/* some special handling before starting the script up */
					/* once we're done then we just re-enter at the generic */
					/* phase handler.  The default case is to enter the */
					/* generic phase handler. */
					case SCSI_DATAI_PHASE:
					{
						/* This function passes back the Current_SG_Entry on the */
						/* the Phase Mismatch because if there was a byte in */
						/* the SWIDE register it will be moved and then interrupt */
						/* The code lower needs to know what the Current SG enry */
						/* is so that it can update the TI entry */
						HandleDataInPM(PCIDevice, &Current_SG_Entry);
						break;
					}
					case SCSI_DATAO_PHASE:
					{
						HandleDataOutPM(PCIDevice);
						break;
					}
					default:                /* currently I can't handle any other phases */
					{
						iowriteDSP(PCIDevice->base_addr2,
							getPhysAddr(rw_script) + Ent_RW_Handle_Phase);
					}
				}       /* end switch on phase mismatch */
			}
			/* all other SCSI interrupts are fatal */
			else
			{
				/* set the major status of the IO */
				rw->IO_major_status = SCSI_ILLEGAL_SCSI_INT;

				/* set the minor status of the IO */
				if (sist & C8XX_INT_CMP) rw->IO_minor_status = SCSI_GOT_CMP_INT;
				if (sist & C8XX_INT_SEL) rw->IO_minor_status = SCSI_GOT_SEL_INT;
				if (sist & C8XX_INT_RSL) rw->IO_minor_status = SCSI_GOT_RSL_INT;
				if (sist & C8XX_INT_SGE) rw->IO_minor_status = SCSI_GOT_SGE_INT;
				if (sist & C8XX_INT_UDC) rw->IO_minor_status = SCSI_GOT_UDC_INT;
				if (sist & C8XX_INT_RST) rw->IO_minor_status = SCSI_GOT_RST_INT;
				if (sist & C8XX_INT_PAR) rw->IO_minor_status = SCSI_GOT_PAR_INT;
				if (sist & C8XX_INT_GEN) rw->IO_minor_status = SCSI_GOT_GEN_INT;
				if (sist & C8XX_INT_HTH) rw->IO_minor_status = SCSI_GOT_HTH_INT;

				/* Check for a complete transfer, unlikely if here */
				if (IORead8(PCIDevice->base_addr2+SBR) & A_RW_ALL_DATA_TRANSFERRED)
					rw->Xfer_status = SCSI_TRANSFER_COMPLETE;
				else
					rw->Xfer_status = SCSI_INCOMPLETE_TRANSFER;

				/* Fail the IO */
				return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector, SCSI_ILLEGAL_INT));
			}
		}
		else if (istat & C8XX_INT_DIP)  /* dma interrupt */
		{
			/* here, the only acceptable interrupt is Script Int Rcvd */
			if (dstat & C8XX_INT_SIR)
			{
				switch(vector)
				{
					case A_RW_Complete_int:
					{
						/* all done */
						/* set the status of the IO */
						rw->IO_major_status = SCSI_COMMAND_COMPLETE;
						rw->IO_minor_status = SCSI_NO_STATUS;

						/* do check condition stuff */
						if (rw->stat_buf == 0x02)
						{
							rw->IO_major_status = SCSI_CHECK_CONDITION;

							if (rw->enable_auto_rqs)
							{
								maint.sel_info = rw->sel_info;
								maint.data_length = 18;
								maint.data_buf = rw->sense_buf;
								sense_ret = do_RqSense(PCIDevice, &maint);

								if (sense_ret == SCSI_GOOD)
									rw->IO_minor_status = SCSI_VALID_SENSE_DATA;
								else
									rw->IO_major_status = SCSI_INVALID_SENSE_DATA;

							}
							else
								rw->IO_minor_status = SCSI_INVALID_SENSE_DATA;
						}

						/* check for all data transferred */
						if (IORead8(PCIDevice->base_addr2+SBR) & A_RW_ALL_DATA_TRANSFERRED)
							rw->Xfer_status = SCSI_TRANSFER_COMPLETE;
						else
							rw->Xfer_status = SCSI_INCOMPLETE_TRANSFER;

						/* if we got a check condition then return a failing */
						/* status */
						if (rw->IO_major_status == SCSI_CHECK_CONDITION)
							return(SCSI_COMMANDFAILED);

						scsi_command_complete = 1;
						break;
					}

					case A_RW_Need_More_SG:
					{
						/* set up more SG entries */
						Fill_SG_Entries(PCIDevice, rw, &NextSGEntry);

						/* now start the script at the data in jump instruction */
						/* start depends on whether this is a read or write */
						if ((rw->cmd_buf[0] & 0x0F) == 0x08)
							/* this is a read so start at data in moves */
							iowriteDSP(PCIDevice->base_addr2,
								getPhysAddr(rw_script) + Ent_RW_Offset_patch_di);
						else
							/* this is a write so start at data out moves */
							iowriteDSP(PCIDevice->base_addr2,
								getPhysAddr(rw_script) + Ent_RW_Offset_patch_do);

						/* That's all, now loop */
						break;
					}

					case A_RW_SWIDE_byte_moved:
					{
						/* update address and count to reflect swide move */
						/* it is assumed we got here from a Phase Mismatch above */
						/* and that the variable Current_SG_Entry is valid */
						rw_buffer_table[RW_DATA_BUF1 + Current_SG_Entry].address++;
						rw_buffer_table[RW_DATA_BUF1 + Current_SG_Entry].count--;
						/* now update jump offset for when reselect happens */
						rw_script[(VINT) (Ent_RW_Offset_patch_di/4) + 1] = Current_SG_Entry * 8;

						/* now start the disconnect script */
						iowriteDSP(PCIDevice->base_addr2,
							getPhysAddr(rw_script) + Ent_RW_Handle_Phase);
						/* now just loop and let the DIP/SIR code handle the int */
						break;
					}

					/* currently I'm going to just crap out on all other interrupts */
					/* as they are generally considered bad.  I could in the future */
					/* try and handle them better, but not now */
					default:
					{
						/* set the major status of the IO */
						rw->IO_major_status = SCSI_ILLEGAL_SCRIPT_INT;

						/* set the minor status of the IO */
						if (vector == A_RW_Not_Tag_Msg)
							rw->IO_minor_status = SCSI_ILLEGAL_TAG_MESSAGE;
						if (vector == A_RW_Unexpected_Message)
							rw->IO_minor_status = SCSI_UNEXPECTED_MESSAGE;
						if (vector == A_RW_Unhandled_Phase_in_bitbucket)
							rw->IO_minor_status = SCSI_BITBUCKET_FAILED;

						/* Check for a complete transfer */
						if (IORead8(PCIDevice->base_addr2+SBR) & A_RW_ALL_DATA_TRANSFERRED)
							rw->Xfer_status = SCSI_TRANSFER_COMPLETE;
						else
							rw->Xfer_status = SCSI_INCOMPLETE_TRANSFER;

						return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector,
												SCSI_ILLEGAL_INT));
					}
				}
			}
			else    /* any other DMA interrupts are fatal, so we'll print a message */
			{               /* and kill the io */

				/* set the major status of the IO */
				rw->IO_major_status = SCSI_ILLEGAL_DMA_INT;

				/* set the minor status of the IO */
				if (sist & C8XX_INT_MDPE) rw->IO_minor_status = SCSI_GOT_MDPE_INT;
				if (sist & C8XX_INT_BF) rw->IO_minor_status = SCSI_GOT_BF_INT;
				if (sist & C8XX_INT_ABRT) rw->IO_minor_status = SCSI_GOT_ABRT_INT;
				if (sist & C8XX_INT_SSI) rw->IO_minor_status = SCSI_GOT_SSI_INT;
				if (sist & C8XX_INT_IID) rw->IO_minor_status = SCSI_GOT_IID_INT;

				/* Check for a complete transfer, unlikely if here */
				if (IORead8(PCIDevice->base_addr2+SBR) & A_RW_ALL_DATA_TRANSFERRED)
					rw->Xfer_status = SCSI_TRANSFER_COMPLETE;
				else
					rw->Xfer_status = SCSI_INCOMPLETE_TRANSFER;

				return(Fail_SCSI(PCIDevice, istat, dstat, sist, vector,
										SCSI_ILLEGAL_INT));
			}
		}
	}

	return(SCSI_GOOD);
}


/******************************************************************************
 Function: Fill_SG_Entries
	Purpose : To set up Scatter Gather entries in the RW script.  The
		  Script can handle up to MAX_SG_ENTRIES scatter gather entries
		  before interrupting.  This routine is called initially to set
		  up up to 16 entries in the Script and is called again
		  for each 16 entries there are from there on.
	Input:  A pointer the pcidev_record and scsirw structures.  And
		a pointer to an integer indicating the next Scatter Gather
		entry to be set up.
	Output: The rw_buffer_table scatter gather entries are initialized for
			the current set of scatter gather entries.
		NextSGEntry is set to the next SG entry number that is not set up
			if more than 16 entries were left to be serviced.
			It is set to -1 if no more entries are to be serviced.
	Assumptions:  None
	Restrictions: None
	Other functions called: RMWon/off to set or clear the bit in the SBR
				register that the script checks to determine
				if there are more scatter gather entries to
				be serviced.

******************************************************************************/

static void Fill_SG_Entries(pcidev_record *PCIDevice, scsirw *rw, VINT *NextSGEntry)

{
	VINT i;
	VINT SGListCount;

	/* there are less than MAX SG entries to be handled */
	if (rw->sg_count-*NextSGEntry <= MAX_SG_ENTRIES)
	{
		SGListCount = *NextSGEntry;
		/* The following loop will set up table indirect entries from */
		/* MAX_SG_ENTRIES - Entries left to be serviced up.  So, if there */
		/* are 5 entries left to be serviced and MAX_SG_ENTRIES is 16 then */
		/* the loop will start at offset 11 in the Table Indirect entry */
		/* list and move up to offset 15 */
		for(i=MAX_SG_ENTRIES-(rw->sg_count-*NextSGEntry); i<MAX_SG_ENTRIES; i++)
		{       /* fill in entries from 16 up */
			rw_buffer_table[RW_DATA_BUF1+i].count = rw->sg_entry[SGListCount].count;
			rw_buffer_table[RW_DATA_BUF1+i].address = rw->sg_entry[SGListCount].address;
			SGListCount++;
		}
		/* patch script so that we jump to the proper move instruction */
		/* first check if we are doing a read or a write as those are handled */
		/* in different spots in the script */
		if ((rw->cmd_buf[0] & 0x0F) == 0x08)
		{
			/* patch read (data in) jump */
			/* The instructions below patches a JUMP REL() script instruction */
			/* that will jump to the proper MOVE instruction in the Scatter */
			/* gather block move list.  Ent_RW_Offset_patch_di must be divided */
			/* by 4 because NASM compiles Entry labels as byte offsets and we */
			/* need a dword offset here.  We add one to get to the second */
			/* dword of the opcode and then set is equal to the relative offset */
			/* of where the first scatter gather move is located.  For example */
			/* if MAX_SG_ENTRIES is 16 and sg_count is 24 and NextSGEntry is */
			/* 16 (meaning we've done the first 16 entries and have just filled */
			/* in the last 8) then the offset would work out to be */
			/* (16 - (24-16)) * 8 or 64 so the JUMP REL() will jump 64 bytes */
			/* relative to the current DSP into the scatter gather move list */
			/* which works out to 16 dwords or 8 Script instructions or halfway */
			/* down the list which is where the first SG move for the final */
			/* 8 entries should be located. */
			rw_script[(VINT) (Ent_RW_Offset_patch_di/4) + 1] =
				(VULONG) (MAX_SG_ENTRIES - (rw->sg_count-*NextSGEntry)) * 8;
		}
		else
		{
			/* patch write (data out) jump */
			/* see the description above for how this formula works */
			rw_script[(VINT) (Ent_RW_Offset_patch_do/4) + 1] =
				(VULONG) (MAX_SG_ENTRIES - (rw->sg_count-*NextSGEntry)) * 8;
		}

		/* set the bit the script checks so that
		/* script knows these are all the SG entries */
		RMWoff(PCIDevice->base_addr2+SBR,
			(VUBYTE) A_RW_NEED_MORE_SG_ENTRIES);
		/* indicate that no more SG entries need to be serviced */
		*NextSGEntry = -1;
	}
	else  /* there are at least MAX SG entries */
	{
		/* since we know that we have at least 16 more entries to */
		/* be service we just start and the first bmove TI entry */
		/* and go all the way to the last one */
		for(i=0; i<MAX_SG_ENTRIES; i++)
		{                       /* fill in all 16 entries */
			rw_buffer_table[RW_DATA_BUF1+i].count =
				rw->sg_entry[*NextSGEntry].count;
			rw_buffer_table[RW_DATA_BUF1+i].address =
				rw->sg_entry[*NextSGEntry].address;
			*NextSGEntry += 1;
		}
		/* patch script so that we jump to the first move instruction */
		/* First see if we are doing a read or write as those are different */
		/* spots in the script to be patched */
		if ((rw->cmd_buf[0] & 0x0F) == 0x08)
		{
			/* patch read (data in) jump */
			/* Since we know we're starting at the beginning of the */
			/* bmove list we just set the jump offset to zero so we */
			/* jump to the first bmove. */
			rw_script[(VINT) (Ent_RW_Offset_patch_di/4) + 1] = 0x00000000;
		}
		else
		{
			/* patch write (data out) jump */
			/* See description above for this formula */
			rw_script[(VINT) (Ent_RW_Offset_patch_do/4) + 1] = 0x00000000;
		}

		/* set the bit the script checks so that script knows we need to give it */
		/* more SG entries */
		RMWon(PCIDevice->base_addr2+SBR, (VUBYTE) A_RW_NEED_MORE_SG_ENTRIES);

		/* from the loop above NextSGEntry indicates the next */
		/* SG entry that needs to be serviced */
	}
}


/******************************************************************************
 Function: HandleDataInPM
	Purpose : To handle clean up after a Phase Mismatch during Data In phase
	Input:  A pointer the pcidev_record and a pointer to a variable which
			  will indicate the Scatter Gather entry that was executing when
			  the PM occured, this is needed by the do_re function if there
			  was a byte in the SWIDE register.
	Output: Current_SG_Entry is filled in with the SG entry that was being
			  serviced.
	Assumptions:  That a phase mismatch has actully occured during data in.
	Restrictions: None
	Other functions called: IORead32 to read chip info
									iowriteDSP to start the script

******************************************************************************/

static void HandleDataInPM(pcidev_record *PCIDevice, VINT *Current_SG_Entry)
{

	VULONG Current_DSP;                     /* Holds current DSP value */

	/* where am I in the SG list? */
	Current_DSP = IORead32(PCIDevice->base_addr2+DSP) - 8;
	*Current_SG_Entry = (VINT) (Current_DSP - FirstDIMove_paddr) / 8;

	/* now update the address and count */
	rw_buffer_table[RW_DATA_BUF1 + *Current_SG_Entry].address +=
		rw_buffer_table[RW_DATA_BUF1 + *Current_SG_Entry].count -
		(IORead32(PCIDevice->base_addr2+DBC) & 0x00FFFFFFl);
	rw_buffer_table[RW_DATA_BUF1 + *Current_SG_Entry].count =
		IORead32(PCIDevice->base_addr2+DBC) & 0x00FFFFFFl;

	/* update the jump offset into the SG list */
	rw_script[(VINT) (Ent_RW_Offset_patch_di/4) + 1] =
		(VULONG) *Current_SG_Entry * 8;

	/* move the byte in SWIDE if necessary */
	if (IORead8(PCIDevice->base_addr2+SCNTL2) & 0x01)
	{
		/* patch move to get byte out of chip */
		rw_script[(VINT) E_RW_Last_di_byte_buf_Used[0]] =
			rw_buffer_table[RW_DATA_BUF1 + *Current_SG_Entry].address;
		/* start script to move byte */
		iowriteDSP(PCIDevice->base_addr2,
			getPhysAddr(rw_script) + Ent_RW_Move_swide_byte);
		/* now just exit and let the upper layer handle the int */
	}
	else    /* nothing in swide so start the disconnect script */
		iowriteDSP(PCIDevice->base_addr2,
			getPhysAddr(rw_script) + Ent_RW_Handle_Phase);
		/* now just exit and let the upper layer handle the int */
}


/******************************************************************************
 Function: HandleDataOutPM
	Purpose : To handle clean up after a Phase Mismatch during Data Out phase
	Input:  A pointer the pcidev_record.
	Output: None
	Assumptions:  That a phase mismatch has actully occured during data out.
	Restrictions: None
	Other functions called: IORead32/8 to read chip info
									RMWon to set bits in chip regiters
									iowriteDSP to start the script

******************************************************************************/

static void HandleDataOutPM(pcidev_record *PCIDevice)
{

	VULONG Current_DSP;                             /* holds current dsp value */
	VINT Current_SG_Entry;                  /* Used to calc. Current SG entry */
	VUINT DFIFO_val;              /* Holds chip DFIFO value */
	VUINT Bytes_remaining;        /* Used to accout for other bytes in chip */

	/* where am I in the SG list? */
	Current_DSP = IORead32(PCIDevice->base_addr2+DSP) - 8;
	Current_SG_Entry = (VINT) (Current_DSP - FirstDOMove_paddr) / 8;

	/* now update the address and count */
	rw_buffer_table[RW_DATA_BUF1 + Current_SG_Entry].address +=
		rw_buffer_table[RW_DATA_BUF1 + Current_SG_Entry].count -
		(IORead32(PCIDevice->base_addr2+DBC) & 0x00FFFFFFl);
	rw_buffer_table[RW_DATA_BUF1 + Current_SG_Entry].count =
		IORead32(PCIDevice->base_addr2+DBC) & 0x00FFFFFFl;

	/* update count and address to reflect any data left in */
	/* the chip */
	/* First check for data in the DMA Fifo */
	/* The variable DFIFO_val is a combination of bits 1-0 of CTEST5 */
	/* and bits 7-0 of the DFIFO register, this will take care */
	/* of both the large (536 byte) fifo devices and all others */
	DFIFO_val = ((IORead8(PCIDevice->base_addr2+CTEST5) & 0x03) << 8) |
					(IORead8(PCIDevice->base_addr2+DFIFO));
	if (IORead8(PCIDevice->base_addr2+CTEST5) & 0x20) /* big fifo */
		Bytes_remaining = (DFIFO_val - (VUINT)
			IORead32(PCIDevice->base_addr2+DBC) & 0x3FF) & 0x3FF;
	else /* little fifo */
		Bytes_remaining = (DFIFO_val - (VUINT)
			IORead32(PCIDevice->base_addr2+DBC) & 0x7F) & 0x7F;

	/* now check the other regs data may be hiding in */
	/* SODL LSB Full?*/
	if (IORead8(PCIDevice->base_addr2+SSTAT0) & 0x20)
		Bytes_remaining++;
	/* SODL MSB Full?*/
	if (IORead8(PCIDevice->base_addr2+SSTAT2) & 0x20)
		Bytes_remaining++;
	/* SODR LSB Full?*/
	if (IORead8(PCIDevice->base_addr2+SSTAT0) & 0x40)
		Bytes_remaining++;
	/* SODR MSB Full?*/
	if (IORead8(PCIDevice->base_addr2+SSTAT2) & 0x40)
		Bytes_remaining++;

	/* Now update the TI entry */
	rw_buffer_table[RW_DATA_BUF1 + Current_SG_Entry].address -=
		Bytes_remaining;
	rw_buffer_table[RW_DATA_BUF1 + Current_SG_Entry].count +=
		Bytes_remaining;

	/* update the jump offset into the SG list */
	rw_script[(VINT) (Ent_RW_Offset_patch_do/4) + 1] =
		(VULONG) Current_SG_Entry * 8;

	/* clear the dma fifo to get any left over data out */
	RMWon(PCIDevice->base_addr2+CTEST3, 0x04);

	/* start the disconnect script */
	iowriteDSP(PCIDevice->base_addr2,
		getPhysAddr(rw_script) + Ent_RW_Handle_Phase);
	/* now just exit and let the upper layer handle the int */

}


/******************************************************************************
 Function: Bitbucket
	Purpose : To clear any addtional bytes from the SCSI bus and a
		  failing io situation.
	Input: A pointer to the pcidev_record structure.
	Output:  If it is possible to clear the scsi bus to a bus free
		 condition then SCSI_COMMANDFAILED is returned.
		 If it is not possible to clear the scsi bus to a bus free
		 condition then SCSI_CON_COMMANDFAILED is returned.
	Assumptions: That scripts are not currently running
	Restrictions: This function can only attempt to clear the scsi bus
				if it is currently in a scsi input phase (i.e. STATUS,
				MSG_IN, or DATA_IN) if it is not in one of those phases
				then this routine cannot attempt to clear the bus.
	Other functions called:  RMWon to clear the fifos
				 IORead8 to check connected status
				 PollISTAT to look for an interrupt from the chip

******************************************************************************/

static VINT Bitbucket(pcidev_record *PCIDevice)

{

	VINT int_return;                        /* Holds return value from PollISTAT */
	VUBYTE dstat, istat;            /* Holds interrupt values from PollISTAT */
	VUINT sist;
	VULONG vector;

	/* first clear all the fifos */
	RMWon(PCIDevice->base_addr2+CTEST3, 0x04);
	RMWon(PCIDevice->base_addr2+STEST3, 0x02);

	/* now check to see if we're currently connected to see if it's worth */
	/* starting the script, this will also prevent a hang condition in the script */
	if (IORead8(PCIDevice->base_addr2+ISTAT) & 0x08)
	{
		/* call script routine to dump data into a bit bucket */
		iowriteDSP(PCIDevice->base_addr2, getPhysAddr(rw_script)
			+ Ent_RW_Bitbucket);
		int_return = PollISTAT( PCIDevice->base_addr2,&vector, &dstat,
										&istat, &sist, 2);
		/* at this point we just want to make sure we got an */
		/* interrupt and that the chip is no longer connected to */
		/* the bus, if yes then return command failed, if no */
		/* return return connected command failed indicating that */
		/* the bus should will most likely need to be reset before */
		/* any more scsi operations can take place */
		if ((int_return == C8XX_GOOD) && !(IORead8(PCIDevice->base_addr2+ISTAT) & 0x08))
			return(SCSI_COMMANDFAILED);
		else
			return(SCSI_CON_COMMANDFAILED);
	}

	else
		return(SCSI_COMMANDFAILED);
}


/******************************************************************************
 Function:  Fail_SCSI
	Purpose : To fail a SCSI io and attempt to clear the scsi bus to a bus
		  free condition
	Input:  A pointer to the pcidev_record data structure.  The interrupt
		registers so that they can be printed out.  The failure type
		which is one of the following:
			SCSI_ILLEGAL_INT - an illegal interrupting condition
			SCSI_POLLTO - a interrupt polling timeout has occured
	Output: Fail_SCSI returns SCSI_COMMANDFAILED if the chip is not
			connected to the SCSI bus upon exit of the routine
		Fail_SCSI returns SCSI_CON_COMMANDFAILED if the chip is
			connected to the SCSI bus upon exit of the routine.
	Assumptions: Scripts are not currently running at the time
	Restrictions: This routine should only be called in catastophic failure
				conditions.
	Other functions called:  BitBucket to attempt to clear the scsi bus to
					bus free.
				 IORead8 to check the connected status of the chip
				 printf to print the error and interrupt regs

******************************************************************************/

static VINT Fail_SCSI(pcidev_record *PCIDevice, VUBYTE istat, VUBYTE dstat,
				 VUINT sist, VULONG vector, VINT fail_type)
{
	/* what type of failure are we looking at */
	switch (fail_type)
	{
		case SCSI_ILLEGAL_INT:
		{
			WriteLog(LOG_PASS, LOG_DETAIL, "SCSI Command Failed...Illegal "
						"interrupting condition\n", NULL);
			sprintf(string, "ISTAT=%02x, DSTAT=%02x, SIST=%04x, VECTOR=%08lx\n",
				istat, dstat, sist, vector);
			WriteLog(LOG_PASS, LOG_DETAIL, string, NULL);
			/* attempt to clear the scsi bus to bus free */
			return(Bitbucket(PCIDevice));
		}
		case SCSI_POLLTO:
		{
			WriteLog(LOG_PASS, LOG_DETAIL, "SCSI Command Failed...Command Timeout\n", NULL);
			/* Since lack of interrupts is the cause of this */
			/* failure it makes no sense to try and clear the */
			/* bus as this requires us to get an interrupt */
			if (!(IORead8(PCIDevice->base_addr2+ISTAT) & 0x08))
				return(SCSI_COMMANDFAILED);
			else
				return(SCSI_CON_COMMANDFAILED);
		}
		default:
		{
			WriteLog(LOG_PASS, LOG_DETAIL, "SCSI Command Failed...Unknown cause\n", NULL);
			if (!(IORead8(PCIDevice->base_addr2+ISTAT) & 0x08))
				return(SCSI_COMMANDFAILED);
			else
				return(SCSI_CON_COMMANDFAILED);
		}
	}
}

