/***************************************************************************
 *                                                                         *
 *  Copyright 1995 Symbios Logic Incorporated.  All rights reserved.       *
 *                                                                         *
 * This software was written by Symbios Logic Inc. to develop and test new *
 * products.  Symbios Logic assumes no liability for its use. This software*
 * is released to the public domain to illustrate certain programming      *
 * techniques for testing the 53c8xx chips                                 *
 *                                                                         *
 ***************************************************************************/

/*           Name:  PCI.C
 *          Title:  Validation test code generic pci tools
 *      $Workfile:  $
 *      $Revision:  $
 *       $Modtime:  $
 *     Programmer:  Symbios Logic
 *  Creation Date:  11/29/95
 *
 *  Version History
 *  ---------------
 *
 *    Date    Who?  Description
 *  --------  ----  -------------------------------------------------------
#BeginRevision
 *  03/06/96  SYM   Formatting changes only.  No functional changes.
#EndRevision
 *
 *  11/29/95  SYM   Initial Revision.
 *
#BeginDescription
 *
 *  This file contains functions which allow access and manipulation of PCI
 *  devices and registers.  It is intended to be as generic as possible
 *  to allow for porting from one hardware interface to another 
 *  (i.e. SCSI to Fibre Channel).  The only thing specific is in the 
 *  PCI_FindSYMDevice routine which assume a Symbios Logic vendor id
 *  of 0x1000.
 *  The core of this code was originally designed by Richard Soloman.
 *
#EndDescription
 *
 *-------------------------------------------------------------------------
 *
 *  $Header: $
 *
 */


#include "valtypes.h"
#include "pci.h"
#include "gen_tool.h"
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>

/* Required for inline code to run through TASM to access 386   */
/* registers used by PCI BIOS 2.0+                              */
#pragma inline

/* Function prototypes */
VINT PCI_FindSYMDevice(pcidev_record *PCIDevice);
void PCI_SetConfigRegister(pcidev_record *PCIDevice, VUINT offset,
							  VULONG value);
VULONG PCI_GetConfigRegister(pcidev_record *PCIDevice, VUINT offset);
VUINT  PCI_GetPCIBIOSVersion(pci_bios_info *pcibios);
void PCI_check_pci_bios(pci_bios_info *pcibios);

/******************************************************************************
 Function: PCI_FindSYMDevice
	Purpose : To determine if a given Symbios Logic device is installed in
				 the PCI system.
	Input: Structure pcidev_record with the following entries filled in:
					device_id       The device id being searched for
					device_index    The index of the device being searched for
	Output: This function fills in the rest of the PCIDevice data structure
				if a device is found and returns PCI_DEVICE_FOUND.
				If no device was found with that ID and INDEX then
				PCI_NO_DEVICE_FOUND is returned with the PCIDevice structure
				unmodified.
	Assumptions: None
	Other functions called: PCI_GetPCIBIOSVersion to make sure there's a PCI BIOS
									PCI_GetConfigRegister to fill in the device rev id

******************************************************************************/

VINT PCI_FindSYMDevice(pcidev_record *PCIDevice)
{

	struct REGPACK regs;    /* register structure required by INTR function */
	VULONG temp_config;     /* temporary holding place for config dword */

	/* make sure there is a PCI bios present */
	if (PCI_GetPCIBIOSVersion(NULL) == PCI_NO_BIOS)
		return(PCI_NO_DEVICE_FOUND);

	/* set up registers for interrupt call */
	regs.r_ax = ((PCI_FUNCTION_ID << 8) | (PCI_FIND_DEVICE));
	regs.r_cx = PCIDevice->device_id;
	regs.r_dx = PCI_SYM_VENDOR_ID;
	regs.r_si = PCIDevice->device_index;
	intr(PCI_BIOS_INT,&regs);

	if (regs.r_flags & 0x01)               /* Carry set, No such device found */
			return(PCI_NO_DEVICE_FOUND);

	/* fill in device structure */
	PCIDevice->bus_num  =  (regs.r_bx & 0xFF00) >> 8; /* BH = bus num */
	PCIDevice->device_num =  (regs.r_bx & 0x00FF);    /* BL = device num/function */
	PCIDevice->function = (regs.r_bx & 0x0007);       /* BL = function */

	/* Command register */
	PCIDevice->command = (VUINT) PCI_GetConfigRegister(PCIDevice,
																			PCI_CONFIG_REG_CMD);
	/* Revision ID */
	PCIDevice->rev_id =  (VUBYTE) PCI_GetConfigRegister(PCIDevice,
																			PCI_CONFIG_REG_REVID);
	/* Cach Line Size and Latency Timer */
	temp_config = PCI_GetConfigRegister(PCIDevice, PCI_CONFIG_REG_CLS);
	PCIDevice->CLS = (VUBYTE) temp_config;
	PCIDevice->latency = (VUBYTE) (temp_config >> 8);

	/* Cardbus CIS ptr */
	PCIDevice->CIS = PCI_GetConfigRegister(PCIDevice, PCI_CONFIG_REG_CIS);

	/* Subsystem ID and Subsystem Vendor ID */
	temp_config = PCI_GetConfigRegister(PCIDevice, PCI_CONFIG_REG_SUBV);
	PCIDevice->sub_ven_id = (VUINT) temp_config;
	PCIDevice->sub_id = (VUINT) (temp_config >> 16);

	/* ROM Base addr */
	PCIDevice->ROM_base_addr = PCI_GetConfigRegister(PCIDevice, PCI_CONFIG_REG_ROM);

	/* Int line and Int pin */
	temp_config = PCI_GetConfigRegister(PCIDevice, PCI_CONFIG_REG_INTL);
	PCIDevice->intl = (VUBYTE) temp_config;
	PCIDevice->intp = (VUBYTE) (temp_config >> 8);


	return(PCI_DEVICE_FOUND);
}



/******************************************************************************
 Function: PCI_SetConfigRegister
	Purpose: To write a particular device's configuration register
	Input: pcidev_record data structure indicating the device to write to
			 offset - the dword offset in config space to write
			 value - the value to write to the config register
	Output: Nothing
	Assumptions: None
	Restrictions: Only dword writes to config space are allowed
	Other functions called: PCI_GetPCIBIOSVersion to make sure there's a PCI BIOS

******************************************************************************/

void PCI_SetConfigRegister(pcidev_record *PCIDevice, VUINT offset,
							  VULONG value)
{

	/* register variables */
	VUINT r_ax, r_bx, r_di, r_dx, r_cx;
	VULONG r_ecx;

	VUINT pci_version;              /* used to store pci bios version */

	r_ax = ((PCI_FUNCTION_ID << 8) | (PCI_WRITE_CONFIG_DWORD));
	r_bx = ((PCIDevice->bus_num & 0xff) << 8) | (PCIDevice->device_num & 0xff);
	r_di = offset;

	pci_version = PCI_GetPCIBIOSVersion(NULL);

	if (pci_version == PCI_BIOS_REV_1X)		/* Version 1.x  */
	{
		r_dx = (VUINT) (value >> 16);
		r_cx = (VUINT) value;
	}
	else if (pci_version == PCI_BIOS_REV_2X)  /* Version 2.x  */
		r_ecx = value;

	else return;          /* oh no!!!!!, no bios */

	/* do PCI function call */
	asm     {
		.386
		mov     ax,[r_ax]
		mov     bx,[r_bx]
		mov     cx,[r_cx]
		mov     dx,[r_dx]
		mov     ecx,DWORD PTR [r_ecx]
		mov     di,[r_di]
		int     PCI_BIOS_INT
	}
}



/******************************************************************************
 Function: PCI_GetConfigRegister
	Purpose: To read a particular devices configuration register
	Input: pcidev_record data structure indicating the device to read from
			 offset - the dword offset in config space to read
	Output: Dword config register value
	Assumptions: None
	Restrictions: Only dword reads to config space are allowed
	Other functions called: PCI_GetPCIBIOSVersion to make sure there's a PCI BIOS

******************************************************************************/

VULONG PCI_GetConfigRegister(pcidev_record *PCIDevice, VUINT offset)
{

	/* register variables */
	VUINT r_ax, r_bx, r_di, r_dx, r_cx;
	VULONG r_ecx;

	VUINT pci_version;      /* used to store pci bios version */
	VULONG result=0;                /* result of PCI Config Read */

	pci_version = PCI_GetPCIBIOSVersion(NULL);

	/* make sure there is a PCI bios present */
	if (pci_version == PCI_NO_BIOS)
		return(0xFFFFFFFFl);

	r_ax = ((PCI_FUNCTION_ID << 8) | (PCI_READ_CONFIG_DWORD));
	r_bx = ((PCIDevice->bus_num & 0xff) << 8) | (PCIDevice->device_num & 0xff);
	r_di = offset;

	asm     {
		.386
		mov     ax,[r_ax]
		mov     bx,[r_bx]
		mov     di,[r_di]
		int     PCI_BIOS_INT
		mov     [r_dx],dx
		mov     [r_cx],cx
		mov     DWORD PTR [r_ecx],ecx
	}

	if (pci_version == PCI_BIOS_REV_1X)       /* Version 1.x  */
	{
		result = r_dx;
		result = (result << 16) | r_cx;
	}
	else if (pci_version == PCI_BIOS_REV_2X)   /* Version 2.x  */
		result = r_ecx;

	else           /* oh no!!!!!, unknown pci bios version */
		result = 0xFFFFFFFFl;

	return(result);
}

/******************************************************************************
 Function: PCI_GetPCIBIOSVersion
	Purpose: To determine what (if any) PCI BIOS version is out there
	Input: pointer to pci_bios_info data structure
			 If this pointer is null then no attempt is made to fill in the
			 structure (can be used for just assuring the existance of a PCI BIOS),
			 if it is a valid pointer then info on the PCI BIOS is returned.
			 This includes the Interface Version, Access Mechanism's supported
			 and the last pci bus in the system.
	Output: PCI_BIOS_REV_1X - PCI BIOS version 1.x found
			  PCI_BIOS_REV_2X - PCI BIOS version 2.x found
			  PCI_UNKNOWN_BIOS - PCI BIOS found but unknown version
			  PCI_NO_BIOS - No PCI BIOS present in the system
			  If pcibios != NULL then BIOS info is filled in and returned in the
			  structure.
	Assumptions: None
	Restrictions: None
	Other functions called: None

******************************************************************************/

VUINT PCI_GetPCIBIOSVersion(pci_bios_info *pcibios)
{
	/* register variables */
	VUINT r_ax, r_bx, r_cx, r_dx;
	VULONG r_edx;

	VULONG sig;             /* PCI Bios Signature string */

	r_ax = ((PCI_FUNCTION_ID << 8) | PCI_BIOS_PRESENT);

	asm     {
		.386
		mov     ax,[r_ax]
		int     PCI_BIOS_INT
		mov     DWORD PTR [r_edx],edx
		mov     [r_dx],dx
		mov     [r_cx],cx
		mov     [r_bx],bx
		mov       [r_ax],ax
	}

	/* make sig = " ICP" */
	sig = ('I' | (' ' << 8));
	sig = ((sig << 16) | ('P' | ('C' << 8)));

	if (r_dx == (('P') | ('C' << 8)))
	{
		if (((r_cx & 0xff) == 'I') && ((r_bx & 0xff00) == 0x0100))
		{
			/* Rev 1.x BIOS       */
			if (pcibios != NULL)
			{
				pcibios->access_mech = r_ax & 0xFF;             /* AL = access mech */
				pcibios->interface_lvl = r_bx;                  /* BX = interface level */
				pcibios->last_bus = r_cx & 0xFF;                        /* CL = last pci bus */
			}
			return(PCI_BIOS_REV_1X);
		}
		else if (r_edx == sig)
		{
			/* Rev 2.x BIOS       */
			if (pcibios != NULL)
			{
				pcibios->access_mech = r_ax & 0xFF;             /* AL = access mech */
				pcibios->interface_lvl = r_bx;                  /* BX = interface level */
				pcibios->last_bus = r_cx & 0xFF;                        /* CL = last pci bus */
			}
			return(PCI_BIOS_REV_2X);
		}
		else
			return(PCI_UNKNOWN_BIOS);
	}
	else
		return(PCI_NO_BIOS);
}

/******************************************************************************
 Function: PCI_check_pci_bios
	Purpose: To check for the existance of a PCI system bios and print
		 out information about that bios.
	Input: pointer to pci_bios_info data structure
	Output: If no PCI bios is found this function prints an error and
		exits the entire program.  If a PCI bios is found then
		the information in the pcibios data structure is filled
		in via a call to PCI_GetPCIBIOSVersion and is printed.
	Assumptions: A valid pointer to the pcibios structure is passed
		     into the function.
	Restrictions: None
	Other functions called: PCI_GetPCIBIOSVersion to determine the 
				existance of a PCI bios and fill in the
				pcibios data structure
				Writelog to output information to a file
				or the screen depending on command line 
				switches
				sprintf format a string for Writelog

******************************************************************************/

void PCI_check_pci_bios(pci_bios_info *pcibios)
{

	VCHAR string[40];       /* Used to create string passed to Writelog */

	/* Check for PCI bios existance */
	if (PCI_GetPCIBIOSVersion(pcibios) == PCI_NO_BIOS)
	{
		WriteLog(LOG_FAIL, LOG_DETAIL, "\nNo PCI Bios found!!! Unable to "
									"continue!!!\n", NULL);
		exit(1);
	}

	/* Output bios information */
	WriteLog(LOG_PASS, LOG_DETAIL, "\nPCI BIOS Found.\n", NULL);
	sprintf(string, "BIOS Version: %02X.%02X\n", ((pcibios->interface_lvl
								& 0xFF00) >> 8), pcibios->interface_lvl & 0x00FF);
	WriteLog(LOG_PASS, LOG_DETAIL, string, NULL);
	sprintf(string, "PCI Buses present: %d\n", pcibios->last_bus+1);
	WriteLog(LOG_PASS, LOG_DETAIL, string, NULL);

	if (pcibios->access_mech & 0x01)
		WriteLog(LOG_PASS, LOG_DETAIL, "Access mechanism #1 supported\n", NULL);

	if (pcibios->access_mech & 0x02)
		WriteLog(LOG_PASS, LOG_DETAIL, "Access mechanism #2 supported\n", NULL);

}


